/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
*                           www.segger.com                           *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

Purpose : embOS CMSIS OS interface

Additional information:
  This embOS CMSIS OS interface supports embOS V5.00 and newer.

  Threads which have their control block and/or stack allocated by the kernel
  and which terminate themself won't free the memory which was allocated by
  the kernel. It's the user's responsibility to either terminate the thread
  by another thread or free the memory after the task was fully terminated.
  OS_TASK_GetID() and OS_STACK_GetTaskStackBase() can be used to retrieve the
  address of the control block and stack before the task terminates itself.
  Another task can then free the memory.

  The deletion of mutexes while a thread has locked them is not supported.

  osThreadDetach() and osThreadJoin() are not supported.

  Message priorities used by message queues are not supported.

  Threads that are created by another thread with the same priority will be
  immediately scheduled.

  Calling osEventFlagsSet(), osEventFlagsClear(), osEventFlagsGet() and
  osEventFlagsWait() from within an ISR is not supported with embOS versions
  V5.16.1 and earlier.

  osKernelSuspend() is only supported in OS_Idle().

*/

#include "cmsis_os2.h"
#include "cmsis_compiler.h"
#include "RTOS.h"

#include <stdlib.h>  // Required for free()

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
#define API_VERSION 20010003  ///< API version (2.1.3)

#if (OS_VERSION >= 51600)
  #if (!defined(OS_InitCalled))
    #define OS_InitCalled  OS_Global.InitCalled
  #endif
#endif

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

// Specifies default stack size for tasks that create their stack dynamically, required by osThreadNew()
#define DEFAULT_STACK_SIZE  (512u)             // Default: 512 Byte

// Specifies maximum period of hardware timer in kernel ticks, required by osKernelSuspend()
#define TIMER_MAX_TICKS     (100u)             // Default: 100 kernel ticks

// Specifies the kernel tick frequency, required by osKernelGetTickFreq()
#define OS_TICK_FREQ        (1000u)            // Default: 1000 ticks per second

// Specifies the timer frequency
#define OS_TIMER_FREQ       (SystemCoreClock)  // Default: SystemCoreClock

/*********************************************************************
*
*       Typedefs
*
**********************************************************************
*/

typedef struct {
  OS_TIMER       Timer;
  const char*    sName;
  osTimerFunc_t  func;
  osTimerType_t  type;
  void*          argument;
  void*          cb_mem_allocated;
} OS_TIMER_NAMED;

typedef struct {
  OS_EVENT       Event;
  const char*    sName;
  void*          cb_mem_allocated;
} OS_EVENT_NAMED;

typedef struct {
  OS_MUTEX       Mutex;
  const char*    sName;
  void*          cb_mem_allocated;
} OS_MUTEX_NAMED;

typedef struct {
  OS_SEMAPHORE   Semaphore;
  const char*    sName;
  uint32_t       max_count;
  void*          cb_mem_allocated;
} OS_SEMAPHORE_NAMED;

typedef struct {
  OS_MEMPOOL     Memf;
  const char*    sName;
  void*          cb_mem_allocated;
  void*          mp_mem_allocated;
} OS_MEMPOOL_NAMED;

typedef struct {
  OS_MAILBOX     Mailbox;
  const char*    sName;
  void*          cb_mem_allocated;
  void*          mq_mem_allocated;
} OS_MAILBOX_NAMED;

typedef struct {
  OS_TASK        Task;
  void*          cb_mem_allocated;     // If not NULL, memory for control block allocated on the heap
  void*          stack_mem_allocated;  // If not NULL, memory for stack allocated on the heap
  osThreadFunc_t pFunc;
} OS_TASK_CMSIS;

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
extern char*    OS_sCopyright;
extern uint32_t SystemCoreClock;

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

static void _ThreadEntry(void* argument) {
  OS_TASK_CMSIS* pTask;

  pTask = (OS_TASK_CMSIS*)OS_TASK_GetID();  // Retrieve actual user thread function from control block
  pTask->pFunc(argument);                   // Call user thread function
  OS_Terminate(NULL);                       // Terminate task properly if user thread function returns. Required for embOS.
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

//  ==== Kernel Management Functions ====

/*********************************************************************
*
*       osKernelInitialize()
*
*  Function description
*    Initialize the RTOS Kernel
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK           : in case of success
*    - osError        : if an unspecific error occurred
*    - osErrorISR     : if called from an Interrupt Service Routine
*    - osErrorNoMemory: if no memory could be reserved for the operation
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*/
osStatus_t osKernelInitialize() {
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osKernelInitialize() from ISR
  }
  if (__get_PRIMASK() != 0u) {
    return osErrorISR;  // Cannot call osKernelInitialize() with disabled interrupts
  }
  if (OS_IsRunning() != 0u) {
    return osError;     // Cannot call osKernelInitialize() after embOS start
  }
  //
  // Initialize embOS
  //
  OS_Init();            // Initialize kernel
  OS_InitHW();          // Initialize hardware for OS
  return osOK;
}

/*********************************************************************
*
*       osKernelGetInfo()
*
*  Function description
*    Get RTOS Kernel Information
*
*  Parameters
*    version: pointer to buffer for retrieving version information
*    id_buf : pointer to buffer for retrieving kernel identification string
*    id_size: size of buffer for kernel identification string
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK      : in case of success
*    - osError   : if an unspecific error occurred
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Allowed to be called from ISRs in CMSIS-RTOS API V2.1.2
*/
osStatus_t osKernelGetInfo(osVersion_t* version, char* id_buf, uint32_t id_size) {
  if (version != NULL) {
    version->api    = API_VERSION;
    version->kernel = OS_VERSION;
  }
  if (id_buf != NULL) {
    strncpy(id_buf, OS_sCopyright, id_size);
  }
  return osOK;
}

/*********************************************************************
*
*       osKernelGetState()
*
*  Function description
*    Get the current RTOS Kernel state
*
*  Return value
*    current RTOS Kernel state
*    - osKernelError    : if an unspecific error occurred or it has been called from an Interrupt Service Routine
*    - osKernelInactive : The kernel is not ready yet. osKernelInitialize() needs to be executed successfully
*    - osKernelReady    : The kernel is not yet running. osKernelStart() transfers the kernel to the running state
*    - osKernelRunning  : The kernel is initialized and running
*    - osKernelLocked   : The kernel was locked with osKernelLock(). The functions osKernelUnlock() or osKernelRestoreLock() unlocks it
*    - osKernelSuspended: The kernel was suspended using osKernelSuspend(). The function osKernelResume() returns to normal operation
*
*  Additional information
*    Renamed from osKernelRunning() in CMSIS-RTOS API V1.2
*    Return type changed from int32_t in CMSIS-RTOS API V1.2
*    Allowed to be called from ISRs in CMSIS-RTOS API V2.1.2
*/
osKernelState_t osKernelGetState() {
#if (OS_DEBUG != 0)
  if (OS_InitCalled == 0u) {
    return osKernelInactive;  // Cannot call osKernelGetState() before kernel initialization
  }
#endif

#if (OS_VERSION < 51600)
  if (OS_Running == 0u) {
#else
  if (OS_Global.IsRunning == 0u) {
#endif
    return osKernelReady;
  }

  if (OS_INT_InInterrupt() != 0u) {
    if (OS_Global.Counters.Cnt.Region > 1u) {
      return osKernelLocked;
    }
  } else {
    if (OS_Global.Counters.Cnt.Region != 0u) {
      return osKernelLocked;
    }
  }

#if (OS_VERSION < 51600)
  if (OS_Global.TicklessFactor != 0) {
#else
  if (OS_Global.TicklessPeriod != 0) {
#endif
    return osKernelSuspended;
  }
  return osKernelRunning;
}

/*********************************************************************
*
*       osKernelStart()
*
*  Function description
*    Start the RTOS Kernel scheduler
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK      : in case of success
*    - osError   : if an unspecific error occurred
*    - osErrorISR: if called from an Interrupt Service Routine
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*/
osStatus_t osKernelStart() {
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // cannot call osKernelStart() from ISR
  }
  if (__get_PRIMASK() != 0u) {
    return osErrorISR;
  }
#if (OS_VERSION < 51600)
  if ((OS_TASK_GetID() != NULL) || (OS_Running != 0u)) {
#else
  if ((OS_TASK_GetID() != NULL) || (OS_Global.IsRunning != 0u)) {
#endif
    return osError;     // cannot call osKernelStart() from Task or more than once
  }
  OS_Start();
  return osOK;          // We should never arrive here
}

/*********************************************************************
*
*       osKernelLock()
*
*  Function description
*    Lock the RTOS Kernel scheduler
*
*  Return value
*    previous lock state
*    - 1         : locked
*    - 0         : not locked
*    - osErrorISR: if called from an Interrupt Service Routine
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
int32_t osKernelLock() {
  int32_t previousState;

  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;       // Cannot call osKernelLock() from ISR
  }
  if (__get_PRIMASK() != 0u) {
    return osErrorISR;
  }
  previousState = (OS_Global.Counters.Cnt.Region == 0u) ? 0 : 1;
  if (previousState == 0) {  // If unlocked...
    OS_TASK_EnterRegion();   // ...lock.
  }
  return previousState;
}

/*********************************************************************
*
*       osKernelUnlock()
*
*  Function description
*    Unlock the RTOS Kernel scheduler
*
*  Return value
*    previous lock state
*    - 1         : locked
*    - 0         : not locked
*    - osErrorISR: if called from an Interrupt Service Routine
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
int32_t osKernelUnlock() {
  int32_t previousState;

  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;       // Cannot call osKernelUnlock() from ISR
  }
  if (__get_PRIMASK() != 0u) {
    return osErrorISR;
  }
  previousState = (OS_Global.Counters.Cnt.Region == 0u) ? 0 : 1;
  if (previousState == 1) {  // If locked...
    OS_TASK_LeaveRegion();   // ...unlock.
  }
  return previousState;
}

/*********************************************************************
*
*       osKernelRestoreLock()
*
*  Function description
*    Restore the RTOS Kernel scheduler lock state
*
*  Parameters
*    lock: lock state obtained by osKernelLock() or osKernelUnlock()
*
*  Return value
*    new lock state
*    - 1         : locked
*    - 0         : not locked
*    - osErrorISR: if called from an Interrupt Service Routine
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
int32_t osKernelRestoreLock(int32_t lock) {
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;                          // Cannot call osKernelRestoreLock() from ISR
  }
  if (lock == 0) {
    if (OS_Global.Counters.Cnt.Region != 0u) {  // If locked...
      OS_TASK_LeaveRegion();                    // ...unlock.
    }
  } else {
    if (OS_Global.Counters.Cnt.Region == 0u) {  // If unlocked...
      OS_TASK_EnterRegion();                    // ...lock.
    }
  }
  return lock;
}

/*********************************************************************
*
*       osKernelSuspend()
*
*  Function description
*    Suspend the RTOS Kernel scheduler
*
*  Return value
*    - time in ticks: for how long the system can sleep or power-down
*    - osErrorISR: if called from an Interrupt Service Routine
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osKernelSuspend() {
  OS_TIME IdleTicks;

  if (OS_INT_InInterrupt() != 0u) {
    return (uint32_t)osErrorISR;                // cannot call osKernelSuspend() from ISR
  }
  OS_INT_IncDI();
  IdleTicks = OS_TICKLESS_GetNumIdleTicks();
  if (IdleTicks > 1) {                          // do not enter sleep mode for durations shorter than a full system tick
    if ((OS_U32)IdleTicks > TIMER_MAX_TICKS) {  // do not enter sleep mode for durations longer than the maximum hardware timer period
      IdleTicks = TIMER_MAX_TICKS;
    }
    OS_TICKLESS_Start(IdleTicks, NULL);
  } else {
    IdleTicks = 0;
  }
  OS_INT_DecRI();
  return (uint32_t)IdleTicks;
}

/*********************************************************************
*
*       osKernelResume()
*
*  Function description
*    Resume the RTOS Kernel scheduler
*
*  Parameters
*    sleep_ticks: time in ticks for how long the system was in sleep or power-down mode
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
void osKernelResume(uint32_t sleep_ticks) {
  OS_TICKLESS_AdjustTime((OS_TIME)sleep_ticks);
}

/*********************************************************************
*
*       osKernelGetTickCount()
*
*  Function description
*    Get the RTOS kernel tick count
*
*  Return value
*    RTOS kernel current tick count
*    - osErrorISR: if called from an Interrupt Service Routine
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Allowed to be called from ISRs in CMSIS-RTOS API V2.1.1
*    Changed kernel tick type to uint32_t in CMSIS-RTOS API V2.1.1
*/
uint32_t osKernelGetTickCount() {
  return OS_TIME_GetTicks();
}

/*********************************************************************
*
*       osKernelGetTickFreq()
*
*  Function description
*    Get the RTOS kernel tick frequency
*
*  Return value
*    frequency of the kernel tick
*    - osErrorISR: if called from an Interrupt Service Routine
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Replaces the macro osKernelSysTickMicroSec in CMSIS-RTOS API V1.2
*    Allowed to be called from ISRs in CMSIS-RTOS API V2.1.1
*/
uint32_t osKernelGetTickFreq() {
  return (uint32_t)OS_TICK_FREQ;
}

/*********************************************************************
*
*       osKernelGetSysTimerCount()
*
*  Function description
*    Get the RTOS kernel system timer count
*
*  Return value
*    RTOS kernel current system timer count as 32-bit value
*
*  Additional information
*    Renamed from osKernelSysTick() in CMSIS-RTOS API V1.2
*/
uint32_t osKernelGetSysTimerCount() {
  // may be called from ISR
  uint32_t t;

#if (OS_VERSION < 50600)
  OS_INT_IncDI();
  t = (uint32_t)OS_GetTime_Cycles();
  OS_INT_DecRI();
#else
  t = (uint32_t)OS_TIME_Get_Cycles();
#endif
  return t;
}

/*********************************************************************
*
*       osKernelGetSysTimerFreq()
*
*  Function description
*    Get the RTOS kernel system timer frequency
*
*  Return value
*    frequency of the system timer
*
*  Additional information
*    New with CMSIS-RTOS API 2.0.
*    May be called from ISR.
*/
uint32_t osKernelGetSysTimerFreq() {
  return OS_TIMER_FREQ;
}

//  ==== Thread Management Functions ====

/*********************************************************************
*
*       osThreadNew()
*
*  Function description
*    Create a thread and add it to Active Threads
*
*  Parameters
*    func    : thread function
*    argument: pointer that is passed to the thread function as start argument
*    attr    : thread attributes; NULL: default values
*
*  Return value
*    thread ID for reference by other functions or NULL in case of error
*
*  Additional information
*    Renamed from osThreadCreate() in CMSIS-RTOS API V1.2
*    Return type changed from osThreadId in CMSIS-RTOS API V1.2
*    Options are now passed using a osThreadAttr_t struct, replacing the osThreadDef macro
*    New function prototype is void func (void *arg) when it was void func (const void *arg) in CMSIS-RTOS API V1.2
*/
osThreadId_t osThreadNew(osThreadFunc_t func, void* argument, const osThreadAttr_t* attr) {
  osThreadAttr_t ThreadAttr;
  OS_TASK_CMSIS* pTask;

  if (OS_INT_InInterrupt() != 0u) {
    return NULL;    // Cannot call osThreadNew() from ISR
  }
#if (OS_DEBUG != 0)
  if (OS_InitCalled == 0u) {
    return NULL;    // Cannot call osThreadNew() before kernel initialization
  }
#endif
  if (__get_PRIMASK() != 0u) {
    return NULL;    // Cannot call osThreadNew() with disabled interrupts
  }
  if (func == NULL) {
    return NULL;    // If no task routine was passed, refuse creation
  }
  if (attr != NULL) {
    if (attr->priority == osPriorityError) {
      return NULL;  // Required by tets suite
    }
    //
    // If attributes were passed, use them
    //
    ThreadAttr = *attr;
  } else {
    //
    // Use default settings
    //
    ThreadAttr.name       = NULL;
    ThreadAttr.attr_bits  = osThreadDetached;
    ThreadAttr.cb_mem     = NULL;
    ThreadAttr.cb_size    = sizeof(OS_TASK_CMSIS);
    ThreadAttr.stack_mem  = NULL;
    ThreadAttr.stack_size = DEFAULT_STACK_SIZE;
    ThreadAttr.priority   = osPriorityNormal;
    ThreadAttr.tz_module  = 0u;
    ThreadAttr.reserved   = 0u;
  }

  if (ThreadAttr.stack_size == 0u) {
    ThreadAttr.stack_size = DEFAULT_STACK_SIZE;
  }
  //
  // Check memory for task control block
  //
  if (ThreadAttr.cb_mem == NULL) {                      // If no TaskCB memory was passed, try allocating TaskCB memory
    pTask = (OS_TASK_CMSIS*)OS_HEAP_malloc(sizeof(OS_TASK_CMSIS));
    if (pTask == NULL) {
      return NULL;                                      // Could not allocate TaskCB memory, refuse creation
    }
    pTask->cb_mem_allocated = (void*)pTask;             // Store address for free() in TCB
  } else {
    if (ThreadAttr.cb_size < sizeof(OS_TASK_CMSIS)) {
      return NULL;                                      // Passed TaskCB memory is insufficient, refuse creation
    }
    pTask = (OS_TASK_CMSIS*)ThreadAttr.cb_mem;
    pTask->cb_mem_allocated = NULL;                     // Mark as not allocated from heap
  }
  //
  // Check memory for task stack
  //
  if (ThreadAttr.stack_mem == NULL) {                   // If no stack area was passed, try allocating stack area
    ThreadAttr.stack_mem = (OS_U8 OS_STACKPTR*)OS_HEAP_malloc(ThreadAttr.stack_size);
    if (ThreadAttr.stack_mem == NULL) {
      if (pTask->cb_mem_allocated != NULL) {
        OS_HEAP_free(pTask->cb_mem_allocated);          // Could not allocate task stack area, free pre-allocated TaskCB memory
      }
      return NULL;                                      // Could not allocate task stack area, refuse creation
    }
    pTask->stack_mem_allocated = ThreadAttr.stack_mem;  // Store address for free() in TCB
  } else {
    pTask->stack_mem_allocated = NULL;                  // Mark as not allocated from heap
  }
  //
  // Create task
  //
  OS_USEPARA(ThreadAttr.attr_bits);                     // Currently not supported with embOS
  OS_USEPARA(ThreadAttr.tz_module);                     // Currently not supported with embOS
  OS_USEPARA(ThreadAttr.reserved);                      // Currently not used (i.e. reserved for future usage)
  pTask->pFunc = func;
  OS_TASK_CreateEx(&pTask->Task, ThreadAttr.name, (unsigned long)ThreadAttr.priority, _ThreadEntry, (OS_U8 OS_STACKPTR*)ThreadAttr.stack_mem, (unsigned int)ThreadAttr.stack_size, 2, argument);
  return (osThreadId_t)pTask;
}

/*********************************************************************
*
*       osThreadGetName()
*
*  Function description
*    Get name of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    name as NULL terminated string
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
const char* osThreadGetName(osThreadId_t thread_id) {
  if (__get_PRIMASK() != 0) {
    return NULL;  // osThreadGetName() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // Cannot call osThreadGetName() from ISR
  }
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u)) {
    return NULL;
  }
#if (OS_TRACKNAME != 0)
  return OS_TASK_GetName((OS_TASK*)thread_id);
#else
  return NULL;    // The requested information is not available in some lib modes
#endif
}

/*********************************************************************
*
*       osThreadGetId()
*
*  Function description
*    Return the thread ID of the current running thread
*
*  Return value
*    thread ID for reference by other functions or NULL in case of error
*
*  Additional information
*    Return type changed from osThreadId in CMSIS-RTOS API V1.2
*    Allowed to be called from ISRs in CMSIS-RTOS API V2.1.3
*/
osThreadId_t osThreadGetId() {
  return (osThreadId_t)OS_TASK_GetID();
}

/*********************************************************************
*
*       osThreadGetState()
*
*  Function description
*    Get current thread state of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    current thread state of the specified thread
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osThreadState_t osThreadGetState(osThreadId_t thread_id) {
  if (__get_PRIMASK() != NULL) {
    return osThreadError;  // osThreadGetState() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osThreadError;  // Cannot call osThreadGetState() from ISR
  }
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u)) {
    return osThreadError;
  }
  if ((OS_TASK*)thread_id == OS_TASK_GetID()) {
    return osThreadRunning;
  }
  if ((((OS_TASK*)thread_id)->Stat & (3uL << 0u)) != 0u) {
    return osThreadBlocked;
  }
  if ((((OS_TASK*)thread_id)->Stat & ((1uL << 2u) | 0xF8uL)) != 0u) {
    return osThreadBlocked;
  }
  return osThreadReady;
}

/*********************************************************************
*
*       osThreadGetStackSize()
*
*  Function description
*    Get stack size of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    stack size in bytes
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osThreadGetStackSize(osThreadId_t thread_id) {
  if (__get_PRIMASK() != NULL) {
    return 0u;  // osThreadGetStackSize() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return 0u;  // cannot call osThreadGetStackSize() from ISR
  }
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u)) {
    return 0u;
  }
#if ((OS_CHECKSTACK != 0) || (OS_SUPPORT_MPU != 0))
  return (uint32_t)OS_STACK_GetTaskStackSize((OS_TASK*)thread_id);
#else
  return 0u;    // The requested information is not available in some lib modes
#endif
}

/*********************************************************************
*
*       osThreadGetStackSpace()
*
*  Function description
*    Get available stack space of a thread based on stack watermark recording during execution
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    remaining stack space in bytes
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osThreadGetStackSpace(osThreadId_t thread_id) {
  if (OS_INT_InInterrupt() != 0u) {
    return 0u;  // Cannot call osThreadGetStackSpace() from ISR
  }
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u)) {
    return 0u;
  }
#if ((OS_CHECKSTACK != 0) || (OS_SUPPORT_MPU != 0))
  return (uint32_t)(OS_STACK_GetTaskStackSize((OS_TASK*)thread_id) - OS_STACK_GetTaskStackUsed((OS_TASK*)thread_id));
#else
  return 0u;    // The requested information is not available in some lib modes
#endif
}

/*********************************************************************
*
*       osThreadSetPriority()
*
*  Function description
*    Change priority of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*    priority : new priority value for the thread function
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the priority of the specified thread has been changed successfully.
*    - osErrorParameter: thread_id is NULL or invalid or priority is incorrect.
*    - osErrorResource:  thread specified by parameter thread_id is in an invalid thread state.
*    - osErrorISR:       the function osThreadSetPriority cannot be called from interrupt service routines.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter types changed from osThreadId and osPriority in CMSIS-RTOS API V1.2
*/
osStatus_t osThreadSetPriority(osThreadId_t thread_id, osPriority_t priority) {
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osThreadSetPriority() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osThreadSetPriority() from ISR
  }
  //
  // Check parameters
  //
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u) ||
      (priority <= osPriorityIdle) || (priority >= osPriorityISR))  {
    return osErrorParameter;
  }
  OS_TASK_SetPriority((OS_TASK*)thread_id, (unsigned long)priority);
  return osOK;
}

/*********************************************************************
*
*       osThreadGetPriority()
*
*  Function description
*    Get current priority of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    current priority value of the specified thread or osPriorityError
*
*  Additional information
*    Return type changed from osPriority in CMSIS-RTOS API V1.2
*/
osPriority_t osThreadGetPriority(osThreadId_t thread_id) {
  OS_TASK* pTask;
  OS_PRIO  Priority;

  pTask = (OS_TASK*)thread_id;
  if (__get_PRIMASK() != 0) {
    return osPriorityError;  // osThreadGetPriority() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osPriorityError;  // Cannot call osThreadGetPriority() from ISR
  }
  //
  // Check parameters
  //
  if ((pTask == NULL) || (OS_TASK_IsTask(pTask) == 0u)) {
    return osPriorityError;
  }
  //
  // Retrieve priority. Get base priority first and if it was
  // successful, check whether the task uses an inherited priority.
  //
  Priority = OS_TASK_GetPriority(pTask);
  if (pTask->Priority > Priority) {
    Priority = pTask->Priority;
  }
  return (osPriority_t)Priority;
}

/*********************************************************************
*
*       osThreadYield()
*
*  Function description
*    Pass control to next thread that is in state READY
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:       control has been passed to the next thread successfully.
*    - osError:    an unspecified error has occurred.
*    - osErrorISR: the function osThreadYield cannot be called from interrupt service routines
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*/
osStatus_t osThreadYield() {
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osThreadYield() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osThreadYield() from ISR
  }
  OS_TASK_Yield();
  return osOK;
}

/*********************************************************************
*
*       osThreadSuspend()
*
*  Function description
*    Suspend execution of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the thread has been suspended successfully.
*    - osErrorParameter: thread_id is NULL or invalid.
*    - osErrorResource:  thread specified by parameter thread_id is in an invalid thread state.
*    - osErrorISR:       the function osThreadSuspend cannot be called from interrupt service routines.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osStatus_t osThreadSuspend(osThreadId_t thread_id) {
  OS_TASK* pTask;

  pTask = (OS_TASK*)thread_id;
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osThreadSuspend() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osThreadSuspend() from ISR
  }
  if ((pTask == NULL) || (OS_TASK_IsTask(pTask) == 0)) {
    return osErrorParameter;
  }
  if (OS_TASK_GetSuspendCnt(pTask) != 0) {
    return osErrorResource;
  }
  OS_TASK_Suspend(pTask);
  return osOK;
}

/*********************************************************************
*
*       osThreadResume()
*
*  Function description
*    Resume execution of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    status code that indicates the execution status of the function
*   - osOK:             the thread has been resumed successfully.
*   - osErrorParameter: thread_id is NULL or invalid.
*   - osErrorResource:  thread specified by parameter thread_id is in an invalid thread state.
*   - osErrorISR:       the function osThreadResume cannot be called from interrupt service routines.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osStatus_t osThreadResume(osThreadId_t thread_id) {
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osThreadResume() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osThreadResume() from ISR
  }
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u)) {
    return osErrorParameter;
  }
  if (OS_TASK_GetSuspendCnt((OS_TASK*)thread_id) == 0) {
    return osErrorResource;
  }
  OS_TASK_Resume((OS_TASK*)thread_id);
  return osOK;
}

/*********************************************************************
*
*       osThreadDetach()
*
*  Function description
*    Detach a thread (thread storage can be reclaimed when thread terminates)
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the attribute of the specified thread has been changed to detached successfully.
*    - osErrorParameter: thread_id is NULL or invalid.
*    - osErrorResource:  thread specified by parameter thread_id is in an invalid thread state.
*    - osErrorISR:       the function osThreadDetach cannot be called from interrupt service routines.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osStatus_t osThreadDetach(osThreadId_t thread_id) {
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;     // Cannot call osThreadDetach() from ISR
  }
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u)) {
    return osErrorParameter;
  }
  return osErrorResource;  // Currently not supported with embOS
}

/*********************************************************************
*
*       osThreadJoin()
*
*  Function description
*    Wait for specified thread to terminate
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             if the thread has already been terminated and joined or once the thread has been terminated and the join operations succeeds.
*    - osErrorParameter: thread_id is NULL or invalid.
*    - osErrorResource:  parameter thread_id is NULL or refers to a thread that is not an active thread or the thread is not joinable.
*    - osErrorISR:       the function osThreadJoin cannot be called from interrupt service routines.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osStatus_t osThreadJoin(osThreadId_t thread_id) {
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;     // Cannot call osThreadJoin() from ISR
  }
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u)) {
    return osErrorParameter;
  }
  return osErrorResource;  // Currently not supported with embOS
}

/*********************************************************************
*
*       osThreadExit()
*
*  Function description
*    Terminate execution of current running thread
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
__NO_RETURN void osThreadExit() {
  OS_TASK* thisTask = OS_TASK_GetID();
  OS_Terminate(thisTask);
  while(1);  // avoid 'noreturn' warning, we should never arrive here
}

/*********************************************************************
*
*       osThreadTerminate()
*
*  Function description
*    Terminate execution of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the specified thread has been removed from the active thread list successfully.
*    - osErrorParameter: thread_id is NULL or invalid.
*    - osErrorResource:  thread specified by parameter thread_id is in an invalid thread state or no other READY thread exists.
*    - osErrorISR:       the function osThreadTerminate cannot be called from interrupt service routines.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osThreadId in CMSIS-RTOS API V1.2
*/
osStatus_t osThreadTerminate(osThreadId_t thread_id) {
  OS_TASK_CMSIS* pTask;

  if (__get_PRIMASK() != 0) {
    return osErrorISR;         // osThreadTerminate() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;         // Cannot call osThreadTerminate() from ISR
  }
  if ((thread_id == NULL) || (OS_TASK_IsTask((OS_TASK*)thread_id) == 0u)) {
    return osErrorParameter;   // Invalid argument
  }

  pTask = (OS_TASK_CMSIS*)thread_id;
  if ((OS_TASK*)thread_id == OS_TASK_GetID()) {
    if (OS_TASK_GetNumTasks() == 1) {
      return osErrorResource;  // Cannot terminate a sole task
    }
  }
  OS_Terminate((OS_TASK*)thread_id);
  //
  // Free automatically allocated memory of another task.
  // Stack has to be freed before TCB, because TCB contains stack address.
  //
  if (pTask->stack_mem_allocated != NULL) {
    OS_HEAP_free(pTask->stack_mem_allocated);
  }
  if (pTask->cb_mem_allocated != NULL) {
    OS_HEAP_free(pTask->cb_mem_allocated);
  }
  return osOK;
}

/*********************************************************************
*
*       osThreadGetCount()
*
*  Function description
*    Get number of active threads
*
*  Return value
*    number of active threads
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osThreadGetCount() {
  OS_TASK* pTask;
  uint32_t Cnt;

  if (__get_PRIMASK() != 0) {
    return 0u;                                // osThreadGetCount() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return 0u;                                // Cannot call osThreadGetCount() from ISR
  }
  OS_INT_IncDI();
  pTask = OS_Global.pTask;
  for (Cnt = 0u; pTask != NULL; Cnt++) {
    if ((pTask->Stat & (3uL << 0u)) != 0u) {  // Count active tasks only (i.e., tasks not in suspended state)
      Cnt--;
    }
    pTask = pTask->pNext;
  }
  OS_INT_DecRI();
  return Cnt;
}

/*********************************************************************
*
*       osThreadEnumerate()
*
*  Function description
*    Enumerate active threads
*
*  Parameters
*    thread_array: pointer to array for retrieving thread IDs
*    array_items:  maximum number of items in array for retrieving thread IDs
*
*  Return value
*    number of enumerated threads
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osThreadEnumerate(osThreadId_t* thread_array, uint32_t array_items) {
  OS_TASK* pTask;
  uint32_t Cnt;

  if (__get_PRIMASK() != 0) {
    return 0;   // osThreadEnumerate() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return 0u;  // Cannot call osThreadEnumerate() from ISR
  }
  OS_INT_IncDI();
  pTask = OS_Global.pTask;
  for (Cnt = 0u; Cnt < array_items; Cnt++) {
    if (pTask != NULL) {
      thread_array[Cnt] = (osThreadId_t*)pTask;
      pTask             = pTask->pNext;
    } else {
      break;
    }
  }
  OS_INT_DecRI();
  return Cnt;
}

//  ==== Thread Flags Functions ====

/*********************************************************************
*
*       osThreadFlagsSet()
*
*  Function description
*    Set the specified Thread Flags of a thread
*
*  Parameters
*    thread_id: thread ID obtained by osThreadNew() or osThreadGetId()
*    flags    : specifies the flags of the thread that shall be set
*
*  Return value
*    thread flags after setting or error code if highest bit set
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osThreadFlagsSet(osThreadId_t thread_id, uint32_t flags) {
  uint32_t r;

  // may be called from ISR
  if ((thread_id == NULL) || (flags & 0x80000000u)) {
    return osFlagsErrorParameter;  // Exceeds the 31 signals available
  }
  OS_INT_IncDI();
  OS_TASKEVENT_Set((OS_TASK*)thread_id, (OS_TASKEVENT)flags);
  r = (uint32_t)(((OS_TASK*)thread_id)->Events);
  OS_INT_DecRI();
  return r;
}

/*********************************************************************
*
*       osThreadFlagsClear()
*
*  Function description
*    Clear the specified Thread Flags of current running thread
*
*  Parameters
*    flags: specifies the flags of the thread that shall be cleared
*
*  Return value
*    thread flags before clearing or error code if highest bit set
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osThreadFlagsClear(uint32_t flags) {
  uint32_t r;

  if (OS_INT_InInterrupt() != 0u) {
    return osFlagsErrorISR;        // Cannot call osThreadFlagsClear() from ISR
  }
  if (flags & 0x80000000u) {
    return osFlagsErrorParameter;  // Exceeds the 31 signals available
  }
  OS_INT_IncDI();
  r = (uint32_t)OS_TASKEVENT_Get(OS_TASK_GetID());
  OS_TASKEVENT_ClearEx(OS_TASK_GetID(), (OS_TASKEVENT)flags);
  OS_INT_DecRI();
  return r;
}

/*********************************************************************
*
*       osThreadFlagsGet()
*
*  Function description
*    Get the current Thread Flags of current running thread
*
*  Return value
*    current thread flags
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osThreadFlagsGet() {
  if (OS_INT_InInterrupt() != 0u) {
    return osFlagsErrorISR;  // Cannot call osThreadFlagsGet() from ISR
  }
  return (uint32_t)OS_TASKEVENT_Get(OS_TASK_GetID());
}

/*********************************************************************
*
*       osThreadFlagsWait()
*
*  Function description
*    Wait for one or more Thread Flags of the current running thread to become signaled
*
*  Parameters
*    flags  : specifies the flags to wait for
*    options: specifies flags options (osFlagsXxxx)
*    timeout: CMSIS_RTOS_TimeOutValue or 0 in case of no time-out
*
*  Return value
*    thread flags before clearing or error code if highest bit set
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osThreadFlagsWait(uint32_t flags, uint32_t options, uint32_t timeout) {
  uint32_t r;
  OS_TIME  t;

  if (OS_INT_InInterrupt() != 0u) {
    return osFlagsErrorISR;                                            // Cannot call osThreadFlagsWait() from ISR
  }
  if (flags & 0x80000000u) {
    return osFlagsErrorParameter;                                      // Exceeds the 31 signals available
  }
  //
  // Do not clear, wait for all
  //
  if ((options & (osFlagsNoClear | osFlagsWaitAll)) == (osFlagsNoClear | osFlagsWaitAll)) {
    if (timeout == 0u) {                                               // wait without timeout
      r = OS_TASKEVENT_Get(NULL);
      if ((r & flags) == flags) {
        return r;                                                      // events (all) set, return mask
      } else {
        return osFlagsErrorResource;                                   // events not (all) set, return error
      }
    } else if (timeout == osWaitForever) {                             // wait indefinitely
      r = 0;
      for (;;) {
        r |= OS_TASKEVENT_Get(NULL);
        if ((r & flags) == flags) {
          return r;                                                    // events (all) set, return mask
        } else {
          OS_TASK_Delay(1);                                            // events not (all) set, suspend task
        }
      }
    } else {                                                           // wait with timeout
      r = 0u;
      t = OS_TIME_GetTicks();
      for (;;) {
        r |= OS_TASKEVENT_Get(NULL);
        if ((r & flags) == flags) {
          return r;                                                    // events (all) set, return mask
        } else if (OS_TIME_GetTicks() >= (t + (OS_TIME)timeout)) {
          return osFlagsErrorTimeout;                                  // events not (all) set, but timeout expired, return error
        } else {
          OS_TASK_Delay(1);                                            // events not (all) set, but timeout endures, suspend task
        }
      }
    }
  //
  // Do not clear, wait for any
  //
  } else if ((options & (osFlagsNoClear)) == osFlagsNoClear) {
    if (timeout == 0u) {                                               // wait without timeout
      r = OS_TASKEVENT_Get(NULL);
      if ((r & flags) != 0u) {
        return r;                                                      // events (any) set, return mask
      } else {
        return osFlagsErrorResource;                                   // events not (any) set, return error
      }
    } else if (timeout == osWaitForever) {                             // wait indefinitely
      r = 0u;
      for (;;) {
        r |= OS_TASKEVENT_Get(NULL);
        if ((r & flags) != 0u) {
          return r;                                                    // events (any) set, return mask
        } else {
          OS_TASK_Delay(1);                                            // events not (any) set, suspend task
        }
      }
    } else {                                                           // wait with timeout
      r = 0u;
      t = OS_TIME_GetTicks();
      for (;;) {
        r |= OS_TASKEVENT_Get(NULL);
        if ((r & flags) != 0u) {
          return r;                                                    // events (any) set, return mask
        } else if (OS_TIME_GetTicks() >= (t + (OS_TIME)timeout)) {
          return osFlagsErrorTimeout;                                  // events not (any) set, but timeout expired, return error
        } else {
          OS_TASK_Delay(1);                                            // events not (any) set, but timeout endures, suspend task
        }
      }
    }
  //
  // Do clear, wait for all
  //
  } else if ((options & osFlagsWaitAll) == osFlagsWaitAll) {
    if (timeout == 0u) {                                               // wait without timeout
      r = OS_TASKEVENT_Get(NULL);
      if ((r & flags) == flags) {
        return OS_TASKEVENT_GetSingleBlocked((OS_TASKEVENT)flags);     // events (all) set, return and clear mask
      } else {
        return osFlagsErrorResource;                                   // events not (all) set, return error
      }
    } else if (timeout == osWaitForever) {                             // wait indefinitely
      r = 0u;
      for (;;) {
        r |= OS_TASKEVENT_Get(NULL);
        if ((r & flags) == flags) {
          return OS_TASKEVENT_GetSingleBlocked((OS_TASKEVENT)flags);   // events (all) set, return and clear mask
        } else {
          OS_TASK_Delay(1);                                            // events not (all) set, suspend task
        }
      }
    } else {                                                           // wait with timeout
      r = 0u;
      t = OS_TIME_GetTicks();
      for (;;) {
        r |= OS_TASKEVENT_Get(NULL);
        if ((r & flags) == flags) {
          return OS_TASKEVENT_GetSingleBlocked((OS_TASKEVENT)flags);   // events (all) set, return and clear mask
        } else if (OS_TIME_GetTicks() >= (t + (OS_TIME)timeout)) {
          return osFlagsErrorTimeout;                                  // events not (all) set, but timeout expired, return error
        } else {
          OS_TASK_Delay(1);                                            // events not (all) set, but timeout endures, suspend task
        }
      }
    }
  //
  // do clear, wait for any
  // clear only the events specified by flags
  // return all signaled events, even those not specified by flags
  //
  } else {
    if (timeout == 0u) {                                               // wait without timeout
      r = OS_TASKEVENT_Get(NULL);
      if ((r & flags) != 0u) {
        return OS_TASKEVENT_ClearEx(NULL, (OS_TASKEVENT)flags);       // events (any) set, return and clear mask
      } else {
        return osFlagsErrorResource;                                   // events not (any) set, return error
      }
    } else if (timeout == osWaitForever) {                             // wait indefinitely
      r = OS_TASKEVENT_GetSingleBlocked((OS_TASKEVENT)flags);
      return (r | OS_TASKEVENT_ClearEx(NULL, (OS_TASKEVENT)flags));
    } else {                                                           // wait with timeout
      r = OS_TASKEVENT_GetSingleTimed((OS_TASKEVENT)flags, (int)timeout);
      if (r == 0) {
        return osFlagsErrorTimeout;
      } else {
        return (r | OS_TASKEVENT_ClearEx(NULL, (OS_TASKEVENT)flags));
      }
    }
  }
}

//  ==== Generic Wait Functions ====

/*********************************************************************
*
*       osDelay()
*
*  Function description
*    Wait for Timeout (Time Delay)
*
*  Parameters
*    ticks: CMSIS_RTOS_TimeOutValue "time ticks" value
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:       the time delay is executed.
*    - osErrorISR: osDelay cannot be called from Interrupt Service Routines.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*/
osStatus_t osDelay(uint32_t ticks) {
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osDelay() from ISR
  }
  if (ticks == 0) {
    return osErrorParameter;
  }
  if (ticks == osWaitForever) {
    while (1) {
      OS_TASK_Delay(0x7FFFFFFF);
    }
  }
  OS_TASK_Delay((OS_TIME)ticks);
  return osOK;
}

/*********************************************************************
*
*       osDelayUntil()
*
*  Function description
*    Wait until specified time
*
*  Parameters
*    ticks: absolute time in ticks
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the time delay is executed.
*    - osErrorParameter: the time is either in the past or cannot be handled (out of bounds).
*    - osErrorISR:       osDelayUntil cannot be called from Interrupt Service Routines.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Changed kernel tick type to uint32_t in CMSIS-RTOS API V2.1.1
*
*/
osStatus_t osDelayUntil(uint32_t ticks) {
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;        // Cannot call osDelayUntil() from ISR
  }
  if ((uint32_t)OS_TIME_GetTicks() >= ticks) {
    return osErrorParameter;  // Cannot wait for time stamp in past
  }
  OS_TASK_DelayUntil((OS_TIME)ticks);
  return osOK;
}

//  ==== Timer Management Functions ====

/*********************************************************************
*
*       _osTimerCallBack()
*
*  Function description
*    embOS Software timers cannot be configured to automatically re-trigger themselves.
*    Hence, this default callback ensures any timer routine will re-trigger itself if necessary.
*/
static void _osTimerCallBack(void) {
  OS_TIMER_NAMED* pTimer;

  pTimer = (OS_TIMER_NAMED*)OS_TIMER_GetCurrent();
  (*(pTimer->func))(pTimer->argument);  // Call user function
  if (pTimer->type == osTimerPeriodic) {
    OS_TIMER_Restart(&pTimer->Timer);   // If timer is periodic, re-trigger it
  }
}

/*********************************************************************
*
*       osTimerNew()
*
*  Function description
*    Create and Initialize a timer
*
*  Parameters
*    func    : start address of a timer call back function
*    type    : osTimerOnce for one-shot or osTimerPeriodic for periodic behavior
*    argument: argument to the timer call back function
*    attr    : timer attributes; NULL: default values
*
*  Return value
*    timer ID for reference by other functions or NULL in case of error
*
*  Additional information
*    Renamed from osTimerCreate() in CMSIS-RTOS API V1.2
*    Return type changed from osTimerId in CMSIS-RTOS API V1.2
*    The parameter list and types have changed
*/
osTimerId_t osTimerNew(osTimerFunc_t func, osTimerType_t type, void* argument, const osTimerAttr_t* attr) {
  osTimerAttr_t   TimerAttr;
  OS_TIMER_NAMED* pTimer;

  if (__get_PRIMASK() != 0) {
    return NULL;                    // osTimerNew() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;                    // Cannot call osTimerNew() from ISR
  }
#if (OS_DEBUG != 0)
  if (OS_InitCalled == 0u) {
    return NULL;                    // Cannot call osTimerNew() before kernel initialization
  }
#endif
  if (func == NULL) {
    return NULL;                    // If no timer routine was passed, refuse creation
  }
  if (attr != NULL) {
    TimerAttr = *attr;              // If attributes were passed, use them
  } else {                          // If no attributes were passed, set default values
    TimerAttr.name      = NULL;
    TimerAttr.attr_bits = 0;
    TimerAttr.cb_mem    = NULL;
    TimerAttr.cb_size   = sizeof(OS_TIMER_NAMED);
  }
  //
  // Check memory for timer control block
  //
  if (TimerAttr.cb_mem == NULL) {   // If no TimerCB memory was passed, try allocating TimerCB memory
    pTimer = OS_HEAP_malloc(sizeof(OS_TIMER_NAMED));
    if (pTimer == NULL) {
      return NULL;                  // Could not allocate TimerCB memory, refuse creation
    }
    pTimer->cb_mem_allocated = pTimer;
  } else if (TimerAttr.cb_size < sizeof(OS_TIMER_NAMED)) {
    return NULL;                    // Passed TimerCB memory is insufficient, refuse creation
  } else {
    pTimer                   = (OS_TIMER_NAMED*)TimerAttr.cb_mem;
    pTimer->cb_mem_allocated = NULL;
  }
  //
  // Create timer
  //
  pTimer->sName    = TimerAttr.name;
  pTimer->func     = func;
  pTimer->type     = type;
  pTimer->argument = argument;
  OS_USEPARA(TimerAttr.attr_bits);  // Currently not used (i.e. reserved for future usage)
  OS_TIMER_Create((OS_TIMER*)pTimer, (OS_TIMERROUTINE*)_osTimerCallBack, (OS_TIME)1);
  return (osTimerId_t)pTimer;
}

/*********************************************************************
*
*       osTimerGetName()
*
*  Function description
*    Get name of a timer
*
*  Parameters
*    timer_id: timer ID obtained by osTimerNew()
*
*  Return value
*    name as NULL terminated string
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
const char* osTimerGetName(osTimerId_t timer_id) {

  if (__get_PRIMASK() != 0) {
    return NULL;  // osTimerGetName() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // cannot call osTimerGetName() from ISR
  }
  if (timer_id == NULL) {
    return NULL;
  }
  return ((OS_TIMER_NAMED*)timer_id)->sName;
}

/*********************************************************************
*
*       osTimerStart()
*
*  Function description
*    Start or restart a timer
*
*  Parameters
*    timer_id: timer ID obtained by osTimerNew()
*    ticks   : CMSIS_RTOS_TimeOutValue "time ticks" value of the timer
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the specified timer has been started or restarted.
*    - osErrorISR:       osTimerStart cannot be called from interrupt service routines.
*    - osErrorParameter: parameter timer_id is either NULL or invalid or ticks is incorrect.
*    - osErrorResource:  the timer specified by parameter timer_id is in an invalid timer state.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osTimerId in CMSIS-RTOS API V1.2
*/
osStatus_t osTimerStart(osTimerId_t timer_id, uint32_t ticks) {
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osTimerStart() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osTimerStart() from ISR
  }
  if ((timer_id == NULL) || (ticks == 0u)) {
    return osErrorParameter;
  }
  if (OS_TIMER_GetStatus((OS_TIMER*)timer_id) == 1u) {
    return osErrorResource;
  }
  OS_TIMER_SetPeriod((OS_TIMER*)timer_id, (int)ticks);
  OS_TIMER_Restart((OS_TIMER*)timer_id);
  return osOK;
}

/*********************************************************************
*
*       osTimerStop()
*
*  Function description
*    Stop a timer
*
*  Parameters
*    timer_id: timer ID obtained by osTimerNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the specified timer has been started or restarted.
*    - osErrorISR:       osTimerStart cannot be called from interrupt service routines.
*    - osErrorParameter: parameter timer_id is either NULL or invalid or ticks is incorrect.
*    - osErrorResource:  the timer specified by parameter timer_id is in an invalid timer state.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osTimerId in CMSIS-RTOS API V1.2
*/
osStatus_t osTimerStop(osTimerId_t timer_id) {
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osTimerStop() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osTimerStop() from ISR
  }
  if (timer_id == NULL) {
    return osErrorParameter;
  }
  if (OS_TIMER_GetStatus((OS_TIMER*)timer_id) == 0u) {
    return osErrorResource;
  }
  OS_TIMER_Stop((OS_TIMER*)timer_id);
  return osOK;
}

/*********************************************************************
*
*       osTimerIsRunning()
*
*  Function description
*    Check if a timer is running
*
*  Parameters
*    timer_id: timer ID obtained by osTimerNew()
*
*  Return value
*    0 not running, 1 running
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osTimerIsRunning(osTimerId_t timer_id) {
  if (OS_INT_InInterrupt() != 0u) {
    return 0u;  // Cannot call osTimerIsRunning() from ISR
  }
  if (timer_id == NULL) {
    return 0u;
  }
  return (uint32_t)OS_TIMER_GetStatus((OS_TIMER*)timer_id);
}

/*********************************************************************
*
*       osTimerDelete()
*
*  Function description
*    Delete a timer
*
*  Parameters
*    timer_id: timer ID obtained by osTimerNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the specified timer has been started or restarted.
*    - osErrorISR:       osTimerStart cannot be called from interrupt service routines.
*    - osErrorParameter: parameter timer_id is either NULL or invalid or ticks is incorrect.
*    - osErrorResource:  the timer specified by parameter timer_id is in an invalid timer state.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osTimerId in CMSIS-RTOS API V1.2
*/
osStatus_t osTimerDelete(osTimerId_t timer_id) {
  void* pTimerFree;

  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osTimerDelete() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osTimerDelete() from ISR
  }
  if (timer_id == NULL) {
    return osErrorParameter;
  }
  pTimerFree = ((OS_TIMER_NAMED*)timer_id)->cb_mem_allocated;
  OS_TIMER_Delete((OS_TIMER*)timer_id);
  if (pTimerFree != NULL) {
    OS_HEAP_free(pTimerFree);
  }
  return osOK;
}

//  ==== Event Flags Management Functions ====

/*********************************************************************
*
*       osEventFlagsNew()
*
*  Function description
*    Create and Initialize an Event Flags object
*
*  Parameters
*    attr: event flags attributes; NULL: default values
*
*  Return value
*    event flags ID for reference by other functions or NULL in case of error
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Event flags and thread flags are replacing the Signal Events in CMSIS-RTOS API V1.2
*/
osEventFlagsId_t osEventFlagsNew(const osEventFlagsAttr_t* attr) {
  osEventFlagsAttr_t EventFlagsAttr;
  OS_EVENT_NAMED*    pEvent;

  if (__get_PRIMASK() != 0) {
    return NULL;                         // osEventFlagsNew() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;                         // Cannot call osEventFlagsNew() from ISR
  }
#if (OS_DEBUG != 0)
  if (OS_InitCalled == 0u) {
    return NULL;                         // Cannot call osEventFlagsNew() before kernel initialization
  }
#endif
  if (attr != NULL) {
    EventFlagsAttr = *attr;              // If attributes were passed, use them
  } else {                               // If no attributes were passed, set default values
    EventFlagsAttr.name      = NULL;
    EventFlagsAttr.attr_bits = 0;
    EventFlagsAttr.cb_mem    = NULL;
    EventFlagsAttr.cb_size   = sizeof(OS_EVENT_NAMED);
  }
  //
  // Check memory for event control block
  //
  if (EventFlagsAttr.cb_mem == NULL) {   // If no EventCB memory was passed, try allocating EventCB memory
    pEvent = OS_HEAP_malloc(sizeof(OS_EVENT_NAMED));
    if (pEvent == NULL) {
      return NULL;                       // Could not allocate EventCB memory, refuse creation
    }
    pEvent->cb_mem_allocated = pEvent;
  } else if (EventFlagsAttr.cb_size < sizeof(OS_EVENT_NAMED)) {
    return NULL;                         // Passed Event memory is insufficient, refuse creation
  } else {
    pEvent                   = (OS_EVENT_NAMED*)EventFlagsAttr.cb_mem;
    pEvent->cb_mem_allocated = NULL;
  }
  //
  // Create event
  //
  pEvent->sName = EventFlagsAttr.name;
  OS_USEPARA(EventFlagsAttr.attr_bits);  // Currently not used (i.e. reserved for future usage)
  OS_EVENT_CreateEx((OS_EVENT*)pEvent, OS_EVENT_RESET_MODE_MANUAL);
  return (osEventFlagsId_t)pEvent;
}

/*********************************************************************
*
*       osEventFlagsGetName()
*
*  Function description
*    Get name of an Event Flags object
*
*  Parameters
*    ef_id: event flags ID obtained by osEventFlagsNew()
*
*  Return value
*    name as NULL terminated string
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Event flags and thread flags are replacing the Signal Events in CMSIS-RTOS API V1.2
*/
const char* osEventFlagsGetName(osEventFlagsId_t ef_id) {
  if (__get_PRIMASK() != 0) {
    return NULL;  // osEventFlagsGetName() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // Cannot call osEventFlagsGetName() from ISR
  }
  if (ef_id == NULL) {
    return NULL;
  }
  return ((OS_EVENT_NAMED*)ef_id)->sName;
}

/*********************************************************************
*
*       osEventFlagsSet()
*
*  Function description
*    Set the specified Event Flags
*
*  Parameters
*    ef_id: event flags ID obtained by osEventFlagsNew()
*    flags: specifies the flags that shall be set
*
*  Return value
*    event flags after setting or error code if highest bit set
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Event flags and thread flags are replacing the Signal Events in CMSIS-RTOS API V1.2
*/
uint32_t osEventFlagsSet(osEventFlagsId_t ef_id, uint32_t flags) {
  uint32_t r;

  if ((ef_id == NULL) || (flags & 0x80000000u)) {
    return osFlagsErrorParameter;                                  // Exceeds the 31 available flags or is invalid
  }
#if (OS_VERSION < 51800)
  if (OS_INT_InInterrupt() != 0u) {
    return osFlagsErrorUnknown;                                    // Cannot call osEventFlagsSet() from ISR with embOS versions V5.16.1 and earlier.
  }
#endif
  OS_INT_IncDI();                                                  // Required to prevent ISRs to alter the mask after calling OS_EVENT_GetMask()
  r  = (uint32_t)OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFFu);  // Get the current event mask
  r |= flags;                                                      // Add the new events to the current event mask which we need to return
  OS_EVENT_SetMask((OS_EVENT*)ef_id, (OS_TASKEVENT)flags);
  OS_INT_DecRI();
  return r;
}

/*********************************************************************
*
*       osEventFlagsClear()
*
*  Function description
*    Clear the specified Event Flags
*
*  Parameters
*    ef_id: event flags ID obtained by osEventFlagsNew()
*    flags: specifies the flags that shall be set
*
*  Return value
*    event flags before clearing or error code if highest bit set
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Event flags and thread flags are replacing the Signal Events in CMSIS-RTOS API V1.2
*/
uint32_t osEventFlagsClear(osEventFlagsId_t ef_id, uint32_t flags) {
  uint32_t r;

  if ((ef_id == NULL) || (flags & 0x80000000u)) {
    return osFlagsErrorParameter;                                       // Exceeds the 31 available flags or is invalid
  }
#if (OS_VERSION < 51800)
  if (OS_INT_InInterrupt() != 0u) {
    return osFlagsErrorUnknown;                                         // Cannot call osEventFlagsClear() from ISR with embOS versions V5.16.1 and earlier.
  }
#endif
  OS_INT_IncDI();
  r = (uint32_t)OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFFu);        // Initially, get signaled events
  OS_EVENT_SetResetMode((OS_EVENT*)ef_id, OS_EVENT_RESET_MODE_AUTO);    // Configure to clear events on OS_EVENT_GetMask()
  OS_EVENT_GetMask((OS_EVENT*)ef_id, (OS_TASKEVENT)flags);              // Now clear specified events
  OS_EVENT_SetResetMode((OS_EVENT*)ef_id, OS_EVENT_RESET_MODE_MANUAL);  // Configure back to manual clearance
  OS_INT_DecRI();
  return r;
}

/*********************************************************************
*
*       osEventFlagsGet()
*
*  Function description
*    Get the current Event Flags
*
*  Parameters
*    ef_id: event flags ID obtained by osEventFlagsNew()
*
*  Return value
*    current event flags
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Event flags and thread flags are replacing the Signal Events in CMSIS-RTOS API V1.2
*/
uint32_t osEventFlagsGet(osEventFlagsId_t ef_id) {
  if (ef_id == NULL) {
    return 0;
  }
#if (OS_VERSION < 51800)
  if (OS_INT_InInterrupt() != 0u) {
    return 0;  // Cannot call osEventFlagsGet() from ISR with embOS versions V5.16.1 and earlier.
  }
#endif
  return (uint32_t)OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFFu);
}

/*********************************************************************
*
*       osEventFlagsWait()
*
*  Function description
*    Wait for one or more Event Flags to become signaled
*
*  Parameters
*    ef_id  : event flags ID obtained by osEventFlagsNew()
*    flags  : specifies the flags to wait for
*    options: specifies flags options (osFlagsXxxx)
*    timeout: CMSIS_RTOS_TimeOutValue or 0 in case of no time-out
*
*  Return value
*    event flags before clearing or error code if highest bit set
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Event flags and thread flags are replacing the Signal Events in CMSIS-RTOS API V1.2
*/
uint32_t osEventFlagsWait(osEventFlagsId_t ef_id, uint32_t flags, uint32_t options, uint32_t timeout) {
  uint32_t r;
  OS_TIME  t;
  OS_TIME  NewTimeout;

  if (timeout != 0) {
    if (__get_PRIMASK() != 0u) {
      return osFlagsErrorParameter;                                                  // May be called with disabled interrupts only if the parameter timeout is set to 0
    }
    if (OS_INT_InInterrupt() != 0u) {
      return osFlagsErrorParameter;                                                  // May be called from Interrupt Service Routines only if the parameter timeout is set to 0
    }
  }
  if ((ef_id == NULL) || (flags & 0x80000000u)) {
    return osFlagsErrorParameter;                                                    // Exceeds the 31 available flags or is invalid
  }
#if (OS_VERSION < 51800)
  if (OS_INT_InInterrupt() != 0u) {
    return osFlagsErrorUnknown;                                                      // Cannot call osEventFlagsWait() from ISR with embOS versions V5.16.1 and earlier.
  }
#endif
  //
  // do not clear, wait for all
  //
  if ((options & (osFlagsNoClear | osFlagsWaitAll)) == (osFlagsNoClear | osFlagsWaitAll)) {
    if (timeout == 0u) {                                                             // wait without timeout
      r = OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      if ((r & flags) == flags) {
        return r;
      } else {
        return osFlagsErrorResource;                                                 // events not (all) set, return error
      }
    } else if (timeout == osWaitForever) {                                           // wait indefinitely
      r = 0;
      while ((r & flags) != flags) {
        r |= OS_EVENT_GetMaskBlocked((OS_EVENT*)ef_id, (OS_TASKEVENT)(~r & flags));  // events (all) set, return and clear mask
      }
      r |= OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      return r;
    } else {                                                                         // wait with timeout
      r          = 0;
      NewTimeout = timeout;
      t          = OS_TIME_GetTicks();
      while ((r & flags) != flags) {
        r |= OS_EVENT_GetMaskTimed((OS_EVENT*)ef_id, (OS_TASKEVENT)(~r & flags), NewTimeout);
        if (r == 0) {
          return osFlagsErrorTimeout;
        }
        NewTimeout = (t + timeout) - OS_TIME_GetTicks();
        if (NewTimeout <= 0) {
          return osFlagsErrorTimeout;
        }
      }
      r |= OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      return r;
    }
  //
  // do not clear, wait for any
  //
  } else if ((options & (osFlagsNoClear)) != 0u) {
    if (timeout == 0u) {                                                             // wait without timeout
      r = OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      if ((r & flags) != 0u) {
        return r;
      } else {
        return osFlagsErrorResource;                                                 // events not (any) set, return error
      }
    } else if (timeout == osWaitForever) {                                           // wait indefinitely
      r  = OS_EVENT_GetMaskBlocked((OS_EVENT*)ef_id, (OS_TASKEVENT)flags);
      r |= OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      return r;
    } else {                                                                         // wait with timeout
      r  = OS_EVENT_GetMaskTimed((OS_EVENT*)ef_id, (OS_TASKEVENT)flags, (int)timeout);
      if (r == 0) {
        return osFlagsErrorTimeout;
      }
      r |= OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      return r;
    }
  //
  // do clear, wait for all
  //
  } else if ((options & osFlagsWaitAll) != 0u) {
    if (timeout == 0u) {                                                             // wait without timeout
      r = OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      if ((r & flags) == flags) {
        OS_INT_IncDI()
        OS_EVENT_Reset((OS_EVENT*)ef_id);
        OS_EVENT_SetMask((OS_EVENT*)ef_id, r & ~flags);
        OS_INT_DecRI();
        return r;
      } else {
        return osFlagsErrorResource;                                                 // events not (all) set, return error
      }
    } else if (timeout == osWaitForever) {                                           // wait indefinitely
      r = 0;
      while ((r & flags) != flags) {
        r |= OS_EVENT_GetMaskBlocked((OS_EVENT*)ef_id, (OS_TASKEVENT)(~r & flags));  // events (all) set, return and clear mask
      }
      OS_INT_IncDI();
      r |= OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      OS_EVENT_Reset((OS_EVENT*)ef_id);
      OS_EVENT_SetMask((OS_EVENT*)ef_id, r & ~flags);
      OS_INT_DecRI();
      return r;
    } else {                                                                         // wait with timeout
      r          = 0;
      NewTimeout = timeout;
      t          = OS_TIME_GetTicks();
      while ((r & flags) != flags) {
        r |= OS_EVENT_GetMaskTimed((OS_EVENT*)ef_id, (OS_TASKEVENT)(~r & flags), NewTimeout);
        if (r == 0) {
          return osFlagsErrorTimeout;
        }
        NewTimeout = (t + timeout) - OS_TIME_GetTicks();
        if (NewTimeout <= 0) {
          return osFlagsErrorTimeout;
        }
      }
      OS_INT_IncDI();
      r |= OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      OS_EVENT_Reset((OS_EVENT*)ef_id);
      OS_EVENT_SetMask((OS_EVENT*)ef_id, r & ~flags);
      OS_INT_DecRI();
      return r;
    }
  //
  // do clear, wait for any
  //
  } else {
    if (timeout == 0u) {                                                             // wait without timeout
      r = OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      if ((r & flags) != 0u) {
        OS_INT_IncDI();
        OS_EVENT_Reset((OS_EVENT*)ef_id);
        OS_EVENT_SetMask((OS_EVENT*)ef_id, r & ~flags);
        OS_INT_DecRI();
        return r;
      } else {
        return osFlagsErrorResource;                                                 // events not (any) set, return error
      }
    } else if (timeout == osWaitForever) {                                           // wait indefinitely
      r  = OS_EVENT_GetMaskBlocked((OS_EVENT*)ef_id, (OS_TASKEVENT)flags);
      OS_INT_IncDI();
      r |= OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      OS_EVENT_Reset((OS_EVENT*)ef_id);
      OS_EVENT_SetMask((OS_EVENT*)ef_id, r & ~flags);
      OS_INT_DecRI();
      return r;
    } else {                                                                         // wait with timeout
      r  = OS_EVENT_GetMaskTimed((OS_EVENT*)ef_id, (OS_TASKEVENT)flags, (int)timeout);
      if (r == 0) {
        return osFlagsErrorTimeout;
      }
      OS_INT_IncDI();
      r |= OS_EVENT_GetMask((OS_EVENT*)ef_id, 0xFFFFFFFF);
      OS_EVENT_Reset((OS_EVENT*)ef_id);
      OS_EVENT_SetMask((OS_EVENT*)ef_id, r & ~flags);
      OS_INT_DecRI();
      return r;
    }
  }
}

/*********************************************************************
*
*       osEventFlagsDelete()
*
*  Function description
*    Delete an Event Flags object
*
*  Parameters
*    ef_id  : event flags ID obtained by osEventFlagsNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK: the specified event flags object has been deleted.
*    - osErrorISR: osEventFlagsDelete cannot be called from interrupt service routines.
*    - osErrorParameter: the value of the parameter ef_id is incorrect.
*    - osErrorResource: parameter ef_id is NULL or wrong.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*    Event flags and thread flags are replacing the Signal Events in CMSIS-RTOS API V1.2
*/
osStatus_t osEventFlagsDelete(osEventFlagsId_t ef_id) {
  void* pEventFree;

  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osEventFlagsDelete() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osEventFlagsDelete() from ISR
  }
  if (ef_id == NULL) {
    return osErrorParameter;
  }
  pEventFree = ((OS_EVENT_NAMED*)ef_id)->cb_mem_allocated;
  OS_EVENT_Delete((OS_EVENT*)ef_id);
  if (pEventFree != NULL) {
    OS_HEAP_free(pEventFree);
  }
  return osOK;
}

//  ==== Mutex Management Functions ====

/*********************************************************************
*
*       osMutexNew()
*
*  Function description
*    Create and Initialize a Mutex object
*
*  Parameters
*    attr: mutex attributes; NULL: default values
*
*  Return value
*    mutex ID for reference by other functions or NULL in case of error
*
*  Additional information
*    Renamed from osMutexCreate() in CMSIS-RTOS API V1.2
*    Return type changed from osMutexId in CMSIS-RTOS API V1.2
*    Parameter type changed from osMutexDef_t in CMSIS-RTOS API V1.2
*/
osMutexId_t osMutexNew(const osMutexAttr_t* attr) {
  osMutexAttr_t   MutexAttr;
  OS_MUTEX_NAMED* pMutex;

  if (__get_PRIMASK() != 0) {
    return NULL;                    // osMutexNew() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;                    // Cannot call osMutexNew() from ISR
  }
#if (OS_DEBUG != 0)
  if (OS_InitCalled == 0u) {
    return NULL;                    // Cannot call osMutexNew() before kernel initialization
  }
#endif
  if (attr != NULL) {
    MutexAttr = *attr;              // If attributes were passed, use them
  } else {                          // If no attributes were passed, set default values
    MutexAttr.name      = NULL;
    MutexAttr.attr_bits = 0;
    MutexAttr.cb_mem    = NULL;
    MutexAttr.cb_size   = sizeof(OS_MUTEX_NAMED);
  }
  //
  // Check memory for mutex control block
  //
  if (MutexAttr.cb_mem == NULL) {   // If no MutexCB memory was passed, try allocating MutexCB memory
    pMutex = OS_HEAP_malloc(sizeof(OS_MUTEX_NAMED));
    if (pMutex == NULL) {
      return NULL;                  // Could not allocate MutexCB memory, refuse creation
    }
    pMutex->cb_mem_allocated = pMutex;
  } else if (MutexAttr.cb_size < sizeof(OS_MUTEX_NAMED)) {
    return NULL;                    // Passed MutexCB memory in insufficient, refuse creation
  } else {
    pMutex                   = (OS_MUTEX_NAMED*)MutexAttr.cb_mem;
    pMutex->cb_mem_allocated = NULL;
  }
  //
  // Create mutex
  //
  pMutex->sName = MutexAttr.name;
  OS_USEPARA(MutexAttr.attr_bits);  // Currently not supported with embOS
  OS_MUTEX_Create((OS_MUTEX*)pMutex);
  return (osMutexId_t)pMutex;
}

/*********************************************************************
*
*       osMutexGetName()
*
*  Function description
*    Get name of a Mutex object
*
*  Parameters
*    mutex_id: mutex ID obtained by osMutexNew()
*
*  Return value
*    name as NULL terminated string
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
const char* osMutexGetName(osMutexId_t mutex_id) {
  if (__get_PRIMASK() != 0) {
    return NULL;  // osMutexGetName() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // Cannot call osMutexGetName() from ISR
  }
  if (mutex_id == NULL) {
    return NULL;
  }
  return ((OS_MUTEX_NAMED*)mutex_id)->sName;
}

/*********************************************************************
*
*       osMutexAcquire()
*
*  Function description
*    Acquire a Mutex or timeout if it is locked
*
*  Parameters
*    mutex_id: mutex ID obtained by osMutexNew()
*    timeout : CMSIS_RTOS_TimeOutValue or 0 in case of no time-out
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the mutex has been obtained.
*    - osErrorTimeout:   the mutex could not be obtained in the given time.
*    - osErrorParameter: parameter mutex_id is NULL or invalid.
*    - osErrorResource:  the mutex specified by parameter mutex_id is in an invalid mutex state or the mutex could not be obtained when no timeout was specified.
*    - osErrorISR:       cannot be called from interrupt service routines.
*
*  Additional information
*    Renamed from osMutexWait() in CMSIS-RTOS API V1.2
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osMutexId in CMSIS-RTOS API V1.2
*/
osStatus_t osMutexAcquire(osMutexId_t mutex_id, uint32_t timeout) {
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osMutexAcquire() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osMutexAcquire() from ISR
  }
  if (mutex_id == NULL) {
    return osErrorParameter;
  }
  if (timeout == 0u) {
    if (OS_MUTEX_Lock((OS_MUTEX*)mutex_id) == 0) {
      return osErrorResource;
    }
  } else if (timeout == osWaitForever) {
    OS_MUTEX_LockBlocked((OS_MUTEX*)mutex_id);
  } else {
    if (OS_MUTEX_LockTimed((OS_MUTEX*)mutex_id, timeout) == 0) {
      return osErrorTimeout;
    }
  }
  return osOK;
}

/*********************************************************************
*
*       osMutexRelease()
*
*  Function description
*    Release a Mutex that was acquired by osMutexAcquire()
*
*  Parameters
*    mutex_id: mutex ID obtained by osMutexNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the mutex has been released.
*    - osErrorParameter: parameter mutex_id is NULL or invalid.
*    - osErrorResource:  the mutex specified by parameter mutex_id is in an invalid mutex state or the mutex was not obtained before/the current thread is not the owner of the mutex.
*    - osErrorISR:       cannot be called from interrupt service routines.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osMutexId in CMSIS-RTOS API V1.2
*/
osStatus_t osMutexRelease(osMutexId_t mutex_id) {
  OS_MUTEX* pMutex;

  pMutex = (OS_MUTEX*)mutex_id;
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osMutexRelease() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osMutexRelease() from ISR
  }
  if (pMutex == NULL) {
    return osErrorParameter;
  }
  if (OS_MUTEX_GetValue(pMutex) == 0) {
    return osErrorResource;
  }
  if ((OS_TASK*)OS_MUTEX_GetOwner(pMutex) != OS_TASK_GetID()) {
    return osErrorResource;
  }
  OS_MUTEX_Unlock(pMutex);
  return osOK;
}

/*********************************************************************
*
*       osMutexGetOwner()
*
*  Function description
*    Get Thread which owns a Mutex object
*
*  Parameters
*    mutex_id: mutex ID obtained by osMutexNew()
*
*  Return value
*    thread ID of owner thread or NULL when mutex was not acquired
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osThreadId_t osMutexGetOwner(osMutexId_t mutex_id) {
  if (__get_PRIMASK() != 0) {
    return NULL;  // osMutexGetOwner() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // Cannot call osMutexGetOwner() from ISR
  }
  if (mutex_id == NULL) {
    return NULL;
  }
  return (osThreadId_t)OS_MUTEX_GetOwner((OS_MUTEX*)mutex_id);
}

/*********************************************************************
*
*       osMutexDelete()
*
*  Function description
*    Delete a Mutex object
*
*  Parameters
*    mutex_id: mutex ID obtained by osMutexNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the mutex has been deleted.
*    - osErrorParameter: parameter mutex_id is NULL or invalid.
*    - osErrorResource:  the mutex specified by parameter mutex_id is in an invalid mutex state.
*    - osErrorISR:       cannot be called from interrupt service routines.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osMutexId in CMSIS-RTOS API V1.2
*/
osStatus_t osMutexDelete(osMutexId_t mutex_id) {
  void* pMutexFree;

  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osMutexDelete() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osMutexDelete() from ISR
  }
  if (mutex_id == NULL) {
    return osErrorParameter;
  }
  if (OS_MUTEX_GetOwner((OS_MUTEX*)mutex_id) != NULL) {
    //
    // Deleting a mutex while a task is holding it is not possible with embOS.
    // Instead of running into OS_Error(), we return osErrorResource.
    //
    return osErrorResource;
  }
  pMutexFree = ((OS_MUTEX_NAMED*)mutex_id)->cb_mem_allocated;
  OS_MUTEX_Delete((OS_MUTEX*)mutex_id);
  if (pMutexFree != NULL) {
    OS_HEAP_free(pMutexFree);
  }
  return osOK;
}

//  ==== Semaphore Management Functions ====

/*********************************************************************
*
*       osSemaphoreNew()
*
*  Function description
*    Create and Initialize a Semaphore object
*
*  Parameters
*    max_count    : maximum number of available tokens
*    initial_count: initial number of available tokens
*    attr         : semaphore attributes; NULL: default values
*
*  Return value
*    semaphore ID for reference by other functions or NULL in case of error
*
*  Additional information
*    Renamed from osSemaphoreCreate() in CMSIS-RTOS API V1.2
*    Return type changed from osSemaphoreId in CMSIS-RTOS API V1.2
*    The parameter list and types have changed
*/
osSemaphoreId_t osSemaphoreNew(uint32_t max_count, uint32_t initial_count, const osSemaphoreAttr_t* attr) {
  osSemaphoreAttr_t   SemaphoreAttr;
  OS_SEMAPHORE_NAMED* pSema;

  if (__get_PRIMASK() != 0) {
    return NULL;                        // osSemaphoreNew() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;                        // Cannot call osSemaphoreNew() from ISR
  }
#if (OS_DEBUG != 0)
  if (OS_InitCalled == 0u) {
    return NULL;                        // Cannot call osSemaphoreNew() before kernel initialization
  }
#endif
  if (attr != NULL) {
    SemaphoreAttr = *attr;              // If attributes were passed, use them
  } else {                              // If no attributes were passed, set default values
    SemaphoreAttr.name      = NULL;
    SemaphoreAttr.attr_bits = 0;
    SemaphoreAttr.cb_mem    = NULL;
    SemaphoreAttr.cb_size   = sizeof(OS_SEMAPHORE_NAMED);
  }
  //
  // Check memory for semaphore control block
  //
  if (SemaphoreAttr.cb_mem == NULL) {   // If no SemaphoreCB memory was passed, try allocating SemaphoreCB memory
    pSema = OS_HEAP_malloc(sizeof(OS_SEMAPHORE_NAMED));
    if (pSema == NULL) {
      return NULL;                      // Could not allocate SemaphoreCB memory, refuse creation
    }
    pSema->cb_mem_allocated = pSema;
  } else if (SemaphoreAttr.cb_size < sizeof(OS_SEMAPHORE_NAMED)) {
    return NULL;                        // Passed SemaphoreCB memory is insufficient, refuse creation
  } else {
    pSema                   = (OS_SEMAPHORE_NAMED*)SemaphoreAttr.cb_mem;
    pSema->cb_mem_allocated = NULL;
  }
  //
  // Create semaphore
  //
  pSema->sName     = SemaphoreAttr.name;
  pSema->max_count = max_count;
  OS_USEPARA(SemaphoreAttr.attr_bits);  // Currently not used (i.e. reserved for future usage)
  OS_SEMAPHORE_Create((OS_SEMAPHORE*)pSema, (unsigned int)initial_count);
  return (osSemaphoreId_t)pSema;
}

/*********************************************************************
*
*       osSemaphoreGetName()
*
*  Function description
*    Get name of a Semaphore object
*
*  Parameters
*    semaphore_id: semaphore ID obtained by osSemaphoreNew()
*
*  Return value
*    name as NULL terminated string
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
const char* osSemaphoreGetName(osSemaphoreId_t semaphore_id) {
  if (__get_PRIMASK() != 0) {
    return NULL;  // osSemaphoreGetName() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // Cannot call osSemaphoreGetName() from ISR
  }
  if (semaphore_id == NULL) {
    return NULL;
  }
  return ((OS_SEMAPHORE_NAMED*)semaphore_id)->sName;
}

/*********************************************************************
*
*       osSemaphoreAcquire()
*
*  Function description
*    Acquire a Semaphore token or timeout if no tokens are available
*
*  Parameters
*    semaphore_id: semaphore ID obtained by osSemaphoreNew()
*    timeout     : CMSIS_RTOS_TimeOutValue or 0 in case of no time-out
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the token has been obtained.
*    - osErrorTimeout:   the token could not be obtained in the given time.
*    - osErrorResource:  the token could not be obtained when no timeout was specified.
*    - osErrorParameter: the parameter semaphore_id is incorrect.
*
*  Additional information
*    Renamed from osSemaphoreWait() in CMSIS-RTOS API V1.2
*    Return type changed from int32_t in CMSIS-RTOS API V1.2
*    Parameter type changed from osSemaphoreId in CMSIS-RTOS API V1.2
*/
osStatus_t osSemaphoreAcquire(osSemaphoreId_t semaphore_id, uint32_t timeout) {
  if (timeout != 0u) {
    if (__get_PRIMASK() != 0u) {
      return osErrorParameter;  // May be called with disabled interrupts only if the parameter timeout is set to 0
    }
    if (OS_INT_InInterrupt() != 0u) {
      return osErrorParameter;  // May be called from Interrupt Service Routines only if the parameter timeout is set to 0
    }
  }
  if (semaphore_id == NULL) {
    return osErrorParameter;
  } else if (timeout == 0u) {
    if (OS_SEMAPHORE_Take((OS_SEMAPHORE*)semaphore_id) == 0u) {
      return osErrorResource;
    }
  } else if (timeout == osWaitForever ) {
    OS_SEMAPHORE_TakeBlocked((OS_SEMAPHORE*)semaphore_id);
  } else {
    if (OS_SEMAPHORE_TakeTimed((OS_SEMAPHORE*)semaphore_id, (int)timeout) == 0) {
      return osErrorTimeout;
    }
  }
  return osOK;
}

/*********************************************************************
*
*       osSemaphoreRelease()
*
*  Function description
*    Release a Semaphore token that was acquired by osSemaphoreAcquire()
*
*  Parameters
*    semaphore_id: semaphore ID obtained by osSemaphoreNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the token has been released.
*    - osErrorResource:  the maximum token count has been reached.
*    - osErrorParameter: the parameter semaphore_id is incorrect.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osSemaphoreId in CMSIS-RTOS API V1.2
*/
osStatus_t osSemaphoreRelease(osSemaphoreId_t semaphore_id) {
  OS_SEMAPHORE_NAMED* pSema;

  // may be called from ISR
  pSema = (OS_SEMAPHORE_NAMED*)semaphore_id;
  if (pSema == NULL) {  // Check semaphore id
    return osErrorParameter;
  }
  if ((uint32_t)OS_SEMAPHORE_GetValue(&pSema->Semaphore) == pSema->max_count) {
    return osErrorResource;
  }
  OS_SEMAPHORE_Give(&pSema->Semaphore);
  return osOK;
}

/*********************************************************************
*
*       osSemaphoreGetCount()
*
*  Function description
*    Get current Semaphore token count
*
*  Parameters
*    semaphore_id: semaphore ID obtained by osSemaphoreNew()
*
*  Return value
*    number of tokens available
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*
*/
uint32_t osSemaphoreGetCount(osSemaphoreId_t semaphore_id) {
  // may be called from ISR
  if (semaphore_id == NULL) {  // Check semaphore id
    return 0u;
  }
  return (uint32_t)OS_SEMAPHORE_GetValue((OS_SEMAPHORE*)semaphore_id);
}

/*********************************************************************
*
*       osSemaphoreDelete()
*
*  Function description
*    Delete a Semaphore object
*
*  Parameters
*    semaphore_id: semaphore ID obtained by osSemaphoreNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the semaphore object has been deleted.
*    - osErrorParameter: the parameter semaphore_id is NULL or invalid.
*    - osErrorResource:  the semaphore specified by parameter semaphore_id is in an invalid semaphore state.
*    - osErrorISR:       osSemaphoreDelete cannot be called from interrupt service routines.
*
*  Additional information
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter type changed from osSemaphoreId in CMSIS-RTOS API V1.2
*/
osStatus_t osSemaphoreDelete(osSemaphoreId_t semaphore_id) {
  void* pSemaFree;

  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osSemaphoreDelete() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osSemaphoreDelete() from ISR
  }
  if (semaphore_id == NULL) {
    return osErrorParameter;
  }
  pSemaFree = ((OS_SEMAPHORE_NAMED*)semaphore_id)->cb_mem_allocated;
  OS_SEMAPHORE_Delete((OS_SEMAPHORE*)semaphore_id);
  if (pSemaFree != NULL) {
    OS_HEAP_free(pSemaFree);
  }
  return osOK;
}

//  ==== Memory Pool Management Functions ====

/*********************************************************************
*
*       osMemoryPoolNew()
*
*  Function description
*    Create and Initialize a Memory Pool object
*
*  Parameters
*    block_count: maximum number of memory blocks in memory pool
*    block_size : memory block size in bytes
*    attr       : memory pool attributes; NULL: default values
*
*  Return value
*    memory pool ID for reference by other functions or NULL in case of error
*
*  Additional information
*    Renamed from osPoolCreate() in CMSIS-RTOS API V1.2
*    Return type changed from osPoolId in CMSIS-RTOS API V1.2
*    Parameter list and parameter types have changed
*/
osMemoryPoolId_t osMemoryPoolNew(uint32_t block_count, uint32_t block_size, const osMemoryPoolAttr_t* attr) {
  osMemoryPoolAttr_t MemoryPoolAttr;
  OS_MEMPOOL_NAMED*  pMemf;

  if (__get_PRIMASK() != 0) {
    return NULL;  // osMemoryPoolNew() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // Cannot call osMemoryPoolNew() from ISR
  }
#if (OS_DEBUG != 0)
  if (OS_InitCalled == 0u) {
    return NULL;  // Cannot call osMemoryPoolNew() before kernel initialization
  }
#endif
  //
  // If basic parameters are invalid, refuse creation
  //
  if ((block_count < 1)          || (block_size < 1) ||
      (block_count > 0x80000000) || (block_size > 0x80000000)) {
    return NULL;
  }
  //
  // Check attributes
  //
  if (attr != NULL) {
    MemoryPoolAttr = *attr;                     // If attributes were passed, use them
  } else {                                      // If no attributes were passed, set default values
    MemoryPoolAttr.name      = NULL;
    MemoryPoolAttr.attr_bits = 0;
    MemoryPoolAttr.cb_mem    = NULL;
    MemoryPoolAttr.cb_size   = sizeof(OS_MEMPOOL_NAMED);
    MemoryPoolAttr.mp_mem    = NULL;
    MemoryPoolAttr.mp_size   = block_count * block_size;
  }
  //
  // Check memory for memory pool control block
  //
  if (MemoryPoolAttr.cb_mem == NULL) {          // If no MemPoolCB memory was passed, try allocating MemPoolCB memory
    pMemf = OS_HEAP_malloc(sizeof(OS_MEMPOOL_NAMED));
    if (pMemf == NULL) {
      return NULL;                              // Could not allocate MemPoolCB memory, refuse creation
    }
    pMemf->cb_mem_allocated = pMemf;
  } else if (MemoryPoolAttr.cb_size < sizeof(OS_MEMPOOL_NAMED)) {
    return NULL;                                // Passed MemPoolCB memory is insufficient, refuse creation
  } else {
    pMemf                   = (OS_MEMPOOL_NAMED*)MemoryPoolAttr.cb_mem;
    pMemf->cb_mem_allocated = NULL;
  }
  //
  // Check memory for memory pool area
  //
  if (MemoryPoolAttr.mp_mem == NULL) {          // If no memory pool area was passed, try allocating memory pool area
    MemoryPoolAttr.mp_mem = OS_HEAP_malloc(block_count * block_size);
    if (MemoryPoolAttr.mp_mem == NULL) {
      if (pMemf->cb_mem_allocated != NULL) {
        OS_HEAP_free(pMemf->cb_mem_allocated);  // Could not allocate memory pool area, free pre-allocated MemPoolCB
      }
      return NULL;                              // Could not allocate memory pool area, refuse creation
    }
    pMemf->mp_mem_allocated = MemoryPoolAttr.mp_mem;
  } else if (MemoryPoolAttr.mp_size < (block_count * block_size)) {
    if (pMemf->cb_mem_allocated != NULL) {
      OS_HEAP_free(pMemf->cb_mem_allocated);    // Could not allocate memory pool area, free pre-allocated MemPoolCB
    }
    return NULL;                                // Passed memory pool area is insufficient, refuse creation
  }
  //
  // Create memory pool
  //
  pMemf->sName = MemoryPoolAttr.name;
  OS_USEPARA(MemoryPoolAttr.attr_bits);         // Currently not used (i.e. reserved for future usage)
  OS_MEMPOOL_Create(&pMemf->Memf, MemoryPoolAttr.mp_mem, block_count, block_size);
  return (osMessageQueueId_t)pMemf;
}

/*********************************************************************
*
*       osMemoryPoolGetName()
*
*  Function description
*    Get name of a Memory Pool object
*
*  Parameters
*    mp_id: memory pool ID obtained by osMemoryPoolNew()
*
*  Return value
*    name as NULL terminated string
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
const char* osMemoryPoolGetName(osMemoryPoolId_t mp_id) {
  if (__get_PRIMASK() != 0) {
    return NULL;  // osMemoryPoolGetName() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // Cannot call osMemoryPoolGetName() from ISR
  }
  if (mp_id == NULL) {
    return NULL;
  }
  return ((OS_MEMPOOL_NAMED*)mp_id)->sName;
}

/*********************************************************************
*
*       osMemoryPoolAlloc()
*
*  Function description
*    Allocate a memory block from a Memory Pool
*
*  Parameters
*    mp_id  : memory pool ID obtained by osMemoryPoolNew()
*    timeout: CMSIS_RTOS_TimeOutValue or 0 in case of no time-out
*
*  Return value
*    address of the allocated memory block or NULL in case of no memory is available
*
*  Additional information
*    Renamed from osPoolAlloc() in CMSIS-RTOS API V1.2
*    Parameter list and parameter types have changed
*/
void* osMemoryPoolAlloc(osMemoryPoolId_t mp_id, uint32_t timeout) {
  OS_MEMPOOL* pMemf;

  pMemf = (OS_MEMPOOL*)mp_id;
  if (timeout != 0u) {
    if (__get_PRIMASK() != 0u) {
      return NULL;  // May be called with disabled interrupts only if the parameter timeout is set to 0
    }
    if (OS_INT_InInterrupt() != 0u) {
      return NULL;  // May be called from Interrupt Service Routines only if the parameter timeout is set to 0
    }
  }
  if (mp_id == NULL) {
    return NULL;
  }
  if (timeout == 0u) {
    return OS_MEMPOOL_Alloc(pMemf);
  }
  if (timeout == osWaitForever) {
    return OS_MEMPOOL_AllocBlocked(pMemf);
  }
  return OS_MEMPOOL_AllocTimed(pMemf, (int)timeout);
}

/*********************************************************************
*
*       osMemoryPoolFree()
*
*  Function description
*    Return an allocated memory block back to a Memory Pool
*
*  Parameters
*    mp_id: memory pool ID obtained by osMemoryPoolNew()
*    block: address of the allocated memory block to be returned to the memory pool
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK: the memory has been freed.
*    - osErrorParameter: parameter mp_id is NULL or invalid, block points to invalid memory.
*    - osErrorResource: the memory pool specified by parameter mp_id is in an invalid memory pool state.
*
*  Additional information
*    Renamed from osPoolFree() in CMSIS-RTOS API V1.2
*    Parameter type changed from osPoolId in CMSIS-RTOS API V1.2
*/
osStatus_t osMemoryPoolFree(osMemoryPoolId_t mp_id, void* block) {
  OS_MEMPOOL* pMemf;

  // may be called from ISR
  pMemf = (OS_MEMPOOL*)mp_id;
  if ((pMemf == NULL) || (block == NULL)) {
    return osErrorParameter;
  }
  if (OS_MEMPOOL_IsInPool(pMemf, block) == 0u) {
    return osErrorResource;
  }
  OS_MEMPOOL_Free(block);
  return osOK;
}

/*********************************************************************
*
*       osMemoryPoolGetCapacity()
*
*  Function description
*    Get maximum number of memory blocks in a Memory Pool
*
*  Parameters
*    mp_id: memory pool ID obtained by osMemoryPoolNew()
*
*  Return value
*    maximum number of memory blocks
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osMemoryPoolGetCapacity(osMemoryPoolId_t mp_id) {
  // may be called from ISR
  if (mp_id == NULL) {
    return 0u;
  }
  return (uint32_t)OS_MEMPOOL_GetNumBlocks((OS_MEMPOOL*)mp_id);
}

/*********************************************************************
*
*       osMemoryPoolGetBlockSize()
*
*  Function description
*    Get memory block size in a Memory Pool
*
*  Parameters
*    mp_id: memory pool ID obtained by osMemoryPoolNew()
*
*  Return value
*    memory block size in bytes
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osMemoryPoolGetBlockSize(osMemoryPoolId_t mp_id) {
  // may be called from ISR
  if (mp_id == NULL) {
    return 0u;
  }
  return (uint32_t)OS_MEMPOOL_GetBlockSize((OS_MEMPOOL*)mp_id);
}

/*********************************************************************
*
*       osMemoryPoolGetCount()
*
*  Function description
*    Get number of memory blocks used in a Memory Pool
*
*  Parameters
*    mp_id: memory pool ID obtained by osMemoryPoolNew()
*
*  Return value
*    number of memory blocks used
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osMemoryPoolGetCount(osMemoryPoolId_t mp_id) {
  // may be called from ISR
  if (mp_id == NULL) {
    return 0u;
  }
  return (uint32_t)(OS_MEMPOOL_GetNumBlocks((OS_MEMPOOL*)mp_id) - OS_MEMPOOL_GetNumFreeBlocks((OS_MEMPOOL*)mp_id));
}

/*********************************************************************
*
*       osMemoryPoolGetSpace()
*
*  Function description
*    Get number of memory blocks available in a Memory Pool
*
*  Parameters
*    mp_id: memory pool ID obtained by osMemoryPoolNew()
*
*  Return value
*    number of memory blocks available
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osMemoryPoolGetSpace(osMemoryPoolId_t mp_id) {
  // may be called from ISR
  if (mp_id == NULL) {
    return 0u;
  }
  return (uint32_t)OS_MEMPOOL_GetNumFreeBlocks((OS_MEMPOOL*)mp_id);
}

/*********************************************************************
*
*       osMemoryPoolDelete()
*
*  Function description
*    Delete a Memory Pool object
*
*  Parameters
*    mp_id: memory pool ID obtained by osMemoryPoolNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK: the memory pool object has been deleted.
*    - osErrorParameter: parameter mp_id is NULL or invalid.
*    - osErrorResource: the memory pool specified by parameter mp_id is in an invalid memory pool state.
*    - osErrorISR: osMemoryPoolDelete cannot be called from interrupt service routines.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osStatus_t osMemoryPoolDelete(osMemoryPoolId_t mp_id) {
  OS_MEMPOOL* pMemf;
  void*       pMemfFree;
  void*       pMemPoolFree;

  pMemf = (OS_MEMPOOL*)mp_id;
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osMemoryPoolDelete() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osMemoryPoolDelete() from ISR
  }
  if (pMemf == NULL) {
    return osErrorParameter;
  }
  if (OS_MEMPOOL_GetNumBlocks(pMemf) != OS_MEMPOOL_GetNumFreeBlocks(pMemf)) {
    return osErrorResource;
  }
  pMemfFree    = (void*)((OS_MEMPOOL_NAMED*)pMemf)->cb_mem_allocated;
  pMemPoolFree = (void*)((OS_MEMPOOL_NAMED*)pMemf)->mp_mem_allocated;
  OS_MEMPOOL_Delete(pMemf);
  if (pMemfFree != NULL) {
    OS_HEAP_free(pMemfFree);
  }
  if (pMemPoolFree != NULL) {
    OS_HEAP_free(pMemPoolFree);
  }
  return osOK;
}

//  ==== Message Queue Management Functions ====

/*********************************************************************
*
*       osMessageQueueNew()
*
*  Function description
*    Create and Initialize a Message Queue object
*
*  Parameters
*    msg_count: maximum number of messages in queue
*    msg_size : maximum message size in bytes
*    attr     : message queue attributes; NULL: default values
*
*  Return value
*    message queue ID for reference by other functions or NULL in case of error
*
*  Additional information
*    Renamed from osMessageCreate() in CMSIS-RTOS API V1.2
*    Return type changed from osMessageQId in CMSIS-RTOS API V1.2
*    Parameter list and parameter types have changed
*/
osMessageQueueId_t osMessageQueueNew(uint32_t msg_count, uint32_t msg_size, const osMessageQueueAttr_t* attr) {
  osMessageQueueAttr_t MessageQueueAttr;
  OS_MAILBOX_NAMED*    pMailbox;

  if (__get_PRIMASK() != 0) {
    return NULL;                                   // osMessageQueueNew() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;                                   // cannot call osMessageQueueNew() from ISR
  }
#if (OS_DEBUG != 0)
  if (OS_InitCalled == 0u) {
    return NULL;                                   // Cannot call osMessageQueueNew() before kernel initialization
  }
#endif
  if ((msg_count < 1)     || (msg_size < 1) ||
      (msg_count > 32767) || (msg_size > 32767)) {
    return NULL;                                   // If basic parameters are invalid, refuse creation
  }
  if (attr != NULL) {
    MessageQueueAttr = *attr;                      // If attributes were passed, use them
  } else {                                         // If no attributes were passed, set default values
    MessageQueueAttr.name      = NULL;
    MessageQueueAttr.attr_bits = 0;
    MessageQueueAttr.cb_mem    = NULL;
    MessageQueueAttr.cb_size   = sizeof(OS_MAILBOX_NAMED);
    MessageQueueAttr.mq_mem    = NULL;
    MessageQueueAttr.mq_size   = msg_size * msg_count;
  }
  //
  // Check memory for mailbox control block
  //
  if (MessageQueueAttr.cb_mem == NULL) {           // If no MailboxCB memory was passed, try allocating MailboxCB memory
    pMailbox = OS_HEAP_malloc(sizeof(OS_MAILBOX_NAMED));
    if (pMailbox == NULL) {
      return NULL;                                 // Could not allocate MailboxCB memory, refuse creation
    }
    pMailbox->cb_mem_allocated = pMailbox;
  } else if (MessageQueueAttr.cb_size < sizeof(OS_MAILBOX_NAMED)) {
    return NULL;                                   // Passed MailboxCB memory is insufficient, refuse creation
  } else {
    pMailbox                   = (OS_MAILBOX_NAMED*)MessageQueueAttr.cb_mem;
    pMailbox->cb_mem_allocated = NULL;             // Mark that cb memory was not allocated from heap
  }
  //
  // Check memory for mail area
  //
  if (MessageQueueAttr.mq_mem == NULL) {           // If no mail area was passed, try allocating memory pool area
    MessageQueueAttr.mq_mem = OS_HEAP_malloc(msg_size * msg_count);
    if (MessageQueueAttr.mq_mem == NULL) {
      if (pMailbox->cb_mem_allocated != NULL) {
        OS_HEAP_free(pMailbox->cb_mem_allocated);  // Could not allocate mail area, free pre-allocated MailboxCB
      }
      return NULL;                                 // Could not allocate mail area, refuse creation
    }
    pMailbox->mq_mem_allocated = MessageQueueAttr.mq_mem;
  } else if (MessageQueueAttr.cb_size < (msg_size * msg_count)) {
    if (pMailbox->cb_mem_allocated != NULL) {
      OS_HEAP_free(pMailbox->cb_mem_allocated);    // Could not allocate mail area, free pre-allocated MailboxCB
    }
    return NULL;                                   // Passed mail area is insufficient, refuse creation
  } else {
    pMailbox->mq_mem_allocated = NULL;             // Mark that mq memory was not allocated from heap
  }
  //
  // Create mailbox
  //
  pMailbox->sName = MessageQueueAttr.name;
  OS_USEPARA(MessageQueueAttr.attr_bits);          // Currently not used (i.e. reserved for future usage)
  OS_MAILBOX_Create((OS_MAILBOX*)pMailbox, (unsigned short)msg_size, (unsigned int)msg_count, MessageQueueAttr.mq_mem);
  return (osMessageQueueId_t)pMailbox;
}

/*********************************************************************
*
*       osMessageQueueGetName()
*
*  Function description
*    Get name of a Message Queue object
*
*  Parameters
*    mq_id: message queue ID obtained by osMessageQueueNew()
*
*  Return value
*    name as NULL terminated string
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
const char* osMessageQueueGetName(osMessageQueueId_t mq_id) {
  if (__get_PRIMASK() != 0) {
    return NULL;  // osMessageQueueGetName() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return NULL;  // Cannot call osMessageQueueGetName() from ISR
  }
  if (mq_id == NULL) {
    return NULL;
  }
  return ((OS_MAILBOX_NAMED*)mq_id)->sName;
}

/*********************************************************************
*
*       osMessageQueuePut()
*
*  Function description
*    Put a Message into a Queue or timeout if Queue is full
*
*  Parameters
*    mq_id   : message queue ID obtained by osMessageQueueNew()
*    msg_ptr : pointer to buffer with message to put into a queue
*    msg_prio: message priority
*    timeout : CMSIS_RTOS_TimeOutValue or 0 in case of no time-out
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the message has been put into the queue.
*    - osErrorTimeout:   the message could not be put into the queue in the given time.
*    - osErrorResource:  not enough space in the queue.
*    - osErrorParameter: the parameter mq_id is incorrect, non-zero timeout specified in an ISR.
*
*  Additional information
*    Renamed from osMessagePut() in CMSIS-RTOS API V1.2
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter list and parameter types have changed
*/
osStatus_t osMessageQueuePut(osMessageQueueId_t mq_id, const void* msg_ptr, uint8_t msg_prio, uint32_t timeout) {
  OS_MAILBOX* pMailbox;

  OS_USEPARA(msg_prio);         // Currently not supported with embOS
  pMailbox = (OS_MAILBOX*)mq_id;
  if (timeout != 0u) {
    if (__get_PRIMASK() != 0u) {
      return osErrorParameter;  // May be called with disabled interrupts only if the parameter timeout is set to 0
    }
    if (OS_INT_InInterrupt() != 0u) {
      return osErrorParameter;  // May be called from Interrupt Service Routines only if the parameter timeout is set to 0
    }
  }
  if (pMailbox == NULL) {
    return osErrorParameter;
  }
  if (timeout == 0u) {
    if (OS_MAILBOX_Put(pMailbox, msg_ptr) != 0) {
      return osErrorResource;
    }
  } else if (timeout == osWaitForever) {
    OS_MAILBOX_PutBlocked(pMailbox, msg_ptr);
  } else {
    if (OS_MAILBOX_PutTimed(pMailbox, msg_ptr, (int)timeout) != 0) {
      return osErrorTimeout;
    }
  }
  return osOK;
}

/*********************************************************************
*
*       osMessageQueueGet()
*
*  Function description
*    Get a Message from a Queue or timeout if Queue is empty
*
*  Parameters
*    mq_id   : message queue ID obtained by osMessageQueueNew()
*    msg_ptr : pointer to buffer for message to get from a queue
*    msg_prio: pointer to buffer for message priority or NULL
*    timeout : CMSIS_RTOS_TimeOutValue or 0 in case of no time-out
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the message has been retrieved from the queue.
*    - osErrorTimeout:   the message could not be retrieved from the queue in the given time.
*    - osErrorResource:  nothing to get from the queue.
*    - osErrorParameter: the parameter mq_id is incorrect, non-zero timeout specified in an ISR.
*
*  Additional information
*    Renamed from osMessageGet() in CMSIS-RTOS API V1.2
*    Return type changed from osStatus in CMSIS-RTOS API V1.2
*    Parameter list and parameter types have changed
*/
osStatus_t osMessageQueueGet(osMessageQueueId_t mq_id, void* msg_ptr, uint8_t* msg_prio, uint32_t timeout) {
  OS_MAILBOX* pMailbox;

  OS_USEPARA(msg_prio);         // Currently not supported with embOS
  pMailbox = (OS_MAILBOX*)mq_id;
  if (timeout != 0) {
    if (__get_PRIMASK() != 0u) {
      return osErrorParameter;  // May be called with disabled interrupts only if the parameter timeout is set to 0
    }
    if (OS_INT_InInterrupt() != 0u) {
      return osErrorParameter;  // May be called from Interrupt Service Routines only if the parameter timeout is set to 0
    }
  }
  if (pMailbox == NULL) {
    return osErrorParameter;
  }
  if (timeout == 0u) {
    if (OS_MAILBOX_Get(pMailbox, msg_ptr) != 0) {
      return osErrorResource;
    }
  } else if (timeout == osWaitForever) {
    OS_MAILBOX_GetBlocked(pMailbox, msg_ptr);
  } else {
    if (OS_MAILBOX_GetTimed(pMailbox, msg_ptr, (int)timeout) != 0) {
      return osErrorTimeout;
    }
  }
  return osOK;
}

/*********************************************************************
*
*       osMessageQueueGetCapacity()
*
*  Function description
*    Get maximum number of messages in a Message Queue
*
*  Parameters
*    mq_id: message queue ID obtained by osMessageQueueNew()
*
*  Return value
*    maximum number of messages
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osMessageQueueGetCapacity(osMessageQueueId_t mq_id) {
  // may be called from ISR
  if (mq_id == NULL) {
    return 0u;
  }
  return (uint32_t)(((OS_MAILBOX*)mq_id)->maxMsg);
}

/*********************************************************************
*
*       osMessageQueueGetMsgSize()
*
*  Function description
*    Get maximum message size in a Message Queue
*
*  Parameters
*    mq_id: message queue ID obtained by osMessageQueueNew()
*
*  Return value
*    maximum message size in bytes
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osMessageQueueGetMsgSize(osMessageQueueId_t mq_id) {
  // may be called from ISR
  if (mq_id == NULL) {
    return 0u;
  }
  return (uint32_t)(((OS_MAILBOX*)mq_id)->sizeofMsg);
}

/*********************************************************************
*
*       osMessageQueueGetCount()
*
*  Function description
*    Get number of queued messages in a Message Queue
*
*  Parameters
*    mq_id: message queue ID obtained by osMessageQueueNew()
*
*  Return value
*    number of queued messages
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osMessageQueueGetCount(osMessageQueueId_t mq_id) {
  // may be called from ISR
  if (mq_id == NULL) {
    return 0u;
  }
  return (uint32_t)OS_MAILBOX_GetMessageCnt((OS_MAILBOX*)mq_id);
}

/*********************************************************************
*
*       osMessageQueueGetSpace()
*
*  Function description
*    Get number of available slots for messages in a Message Queue
*
*  Parameters
*    mq_id: message queue ID obtained by osMessageQueueNew()
*
*  Return value
*    number of available slots for messages
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
uint32_t osMessageQueueGetSpace(osMessageQueueId_t mq_id) {
  // may be called from ISR
  if (mq_id == NULL) {
    return 0u;
  }
  return (uint32_t)(((OS_MAILBOX*)mq_id)->maxMsg - OS_MAILBOX_GetMessageCnt((OS_MAILBOX*)mq_id));
}

/*********************************************************************
*
*       osMessageQueueReset()
*
*  Function description
*    Reset a Message Queue to initial empty state
*
*  Parameters
*    mq_id: message queue ID obtained by osMessageQueueNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the message queue has been reset.
*    - osErrorParameter: parameter mq_id is NULL or invalid.
*    - osErrorResource:  the message queue specified by parameter mq_id is in an invalid message queue state.
*    - osErrorISR:       osMessageQueueReset cannot be called from interrupt service routines.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osStatus_t osMessageQueueReset(osMessageQueueId_t mq_id) {
  if (__get_PRIMASK() != 0) {
    return osErrorISR;  // osMessageQueueReset() may not be called with set PRIMASK
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osMessageQueueReset() from ISR
  }
  if (mq_id == NULL) {
    return osErrorParameter;
  }
  OS_MAILBOX_Clear((OS_MAILBOX*)mq_id);
  return osOK;
}

/*********************************************************************
*
*       osMessageQueueDelete()
*
*  Function description
*    Delete a Message Queue object
*
*  Parameters
*    mq_id: message queue ID obtained by osMessageQueueNew()
*
*  Return value
*    status code that indicates the execution status of the function
*    - osOK:             the message queue object has been deleted.
*    - osErrorParameter: parameter mq_id is NULL or invalid.
*    - osErrorResource:  the message queue specified by parameter mq_id is in an invalid message queue state.
*    - osErrorISR:       osMessageQueueDelete cannot be called from interrupt service routines.
*
*  Additional information
*    New with CMSIS-RTOS API 2.0
*/
osStatus_t osMessageQueueDelete(osMessageQueueId_t mq_id) {
  OS_MAILBOX* pMailbox;
  void*       pMailboxFree;
  void*       pBufferFree;

  pMailbox = (OS_MAILBOX*)mq_id;
  if (__get_PRIMASK() != 0) {
    return osErrorISR;
  }
  if (OS_INT_InInterrupt() != 0u) {
    return osErrorISR;  // Cannot call osMessageQueueDelete() from ISR
  }
  if (pMailbox == NULL) {
    return osErrorParameter;
  }
  pMailboxFree = ((OS_MAILBOX_NAMED*)pMailbox)->cb_mem_allocated;
  pBufferFree  = ((OS_MAILBOX_NAMED*)pMailbox)->mq_mem_allocated;
  OS_MAILBOX_Delete(pMailbox);
  if (pMailboxFree != NULL) {
    OS_HEAP_free(pMailboxFree);
  }
  if (pBufferFree != NULL) {
    OS_HEAP_free(pBufferFree);
  }
  return osOK;
}

/*************************** End of file ****************************/
